"""Parser to obtain a dictionary of key-value pairs from an Onshape model's URL
and add those variables to an RST file.

Relies on the Onshape Documenter feature:
https://cad.onshape.com/documents/6b5c9b74e331c4d03a7c6b01/w/6f98333f14625dd1bdcac2f7/e/35b3d3018f18ec53eeecded7
"""

import json
import math
import os
from shutil import copyfile
from onshape_client.oas import BTFeatureScriptEvalCall2377
from onshape_client.onshape_url import OnshapeElement
from onshape_client import Client
from aguaclara.core.units import u

ureg =  u

msg_str = "message"
val_str = "value"
key_str = "key"

# create global roles using this: https://stackoverflow.com/questions/9698702/how-do-i-create-a-global-role-roles-in-sphinx
# If this grows too much, we'll need to add a global rst as described in the post above.
def parse_quantity(q, for_docs=True):
    """Parse an Onshape units definition

    Args:
        q: an Onshape units definition... for instance:
            {
              'typeTag': '',
              'unitToPower': [
                {
                  'key': 'METER',
                  'value': 1
                }
              ],
              'value': 0.0868175271040671
            }
        for_docs: True if parsing variables for AIDE documentation,
            False otherwise (e.g. validation)

    Returns:
        a string that can be converted to any other unit engine.
    """
    units_s = q[val_str]
    for unit in q["unitToPower"]:
        units_s = units_s * ureg(unit[key_str].lower()) ** unit[val_str]
        try:
            log = math.floor(math.log10(units_s.magnitude))
        except:
            log = 0
        if for_docs:
            if unit[key_str] == 'METER' and unit[val_str] == 1:
                if log >= 3:
                    units_s = units_s.to(ureg.kilometer)
                elif log >= -2 and log <= -1:
                    units_s = units_s.to(ureg.centimeter)
                elif log <= -3:
                    units_s = units_s.to(ureg.millimeter)
            elif unit[key_str] == 'METER' and unit[val_str] == 2:
                if log >= 6:
                    units_s = units_s.to(ureg.kilometer**2)
                elif log >= -4 and log <= -1:
                    units_s = units_s.to(ureg.centimeter**2)
                elif log <= -5:
                    units_s = units_s.to(ureg.millimeter**2)
            elif unit[key_str] == 'METER' and unit[val_str] == 3:
                log += 3
                if log >= 3:
                    units_s = units_s.to(ureg.kiloliter)
                elif log <= -1:
                    units_s = units_s.to(ureg.milliliter)
                else:
                    units_s = units_s.to(ureg.liter)
            return f'{round(units_s, 2):~}'
        else:
            return units_s

def is_fs_type(candidate, type_name):
    """Checks if the a JSON entry is of a specific FeatureScript type.

    Args:
        candidate: decoded JSON object to check the type of
        type_name: string of the FeatureScript Type to check for

    Returns:
        result: True if candidate is of type_name, False otherwise
    """
    result = False
    try:
        if isinstance(type_name, str):
            result = type_name == candidate["typeName"]
        elif isinstance(type_name, list):
            result = any(
                [type_name_one == candidate["typeName"] for type_name_one in type_name]
            )
    except Exception:
        result = False
    return result

def copy_to_docs(file_path, new_name=None, base="doc_files"):
    """First, searches recursively searches for the base path in parent folders.
    Then copies a file to the current working directory. The new file's path
    will be identical to the old file's path relative to the base path.

    Args:
        file_path: path to the file to be copied
        base: base path to use in creating relative file path of the copy
        new_name: new name for the file to avoid duplication.
            Default: None, use existing name

    Returns:
        none
    """
    dir = os.getcwd()
    while not os.path.exists(os.path.join(dir, base)):
        dir = os.path.dirname(dir)

    basepath = os.path.join(dir, base)
    new_path = new_name if new_name is not None else file_path
    try:
        copyfile(os.path.join(basepath, file_path), new_path)
    except IOError as io_err:
        os.makedirs(os.path.dirname(new_path))
        copyfile(os.path.join(basepath, file_path), new_path)

def parse_variables_from_list(unparsed, for_docs=True):
    """Helper function for parse_variables_from_map parses values from a list
    instead of a map.

    Args:
        unparsed: portion of deserialized JSON which has yet to be parsed
        for_docs: True if parsing variables for AIDE documentation,
            False otherwise (e.g. validation)

    Returns:
        measurement_list: list of parsed values
    """
    measurement_list = []

    for to_parse in unparsed:
        if is_fs_type(to_parse, "BTFSValueWithUnits"):
            measurement_list.append(parse_quantity(to_parse[msg_str], for_docs))
        elif is_fs_type(to_parse, ["BTFSValueNumber", "BTFSValueString"]):
            measurement_list.append(to_parse[msg_str][val_str])

    return measurement_list

def merge_index_sections(new_section, old_section):
    """Helper function for merge_indexes which loops through each section and
    combines them.

    Args:
        new_section: section which is being added to if line from old_section is absent
        old_section: section which is pulled from

    Returns:
        none
    """
    for line in old_section:
        if line in new_section:
            continue
        else:
            new_section.append(line)

    return new_section

def find_index_section_limits(filename, section_start=".. toctree::\n",
                              section_end="\n"):
    """Helper function for merge_indexes which loops through the
    file and marks the beginning and end of each section.

    Args:
        filename: path to file to be modified
        section_start: string which marks the start of each section
            Default: '.. toctree::\n'
        section_end: string which marks the end of each section
            Default: '\n'

    Returns:
        lines: list of strings of each line in the file
        section_limits: list of the form [[start1, end1], [start2, end2]]
            which marks the separation between sections
    """
    section_limits = []
    start = 0
    first_newline = True
    index_file = open(filename, "r+")
    lines = index_file.readlines()
    index_file.close()

    for i, line in enumerate(lines):
        if line == section_start:
            start = i
        if line == section_end and start != 0:
            if first_newline:
                first_newline = False
            else:
                end = i
                section_limits.append([start, end])
                start = end = 0
                first_newline = True

    return lines, section_limits

def merge_indexes(new_index, old_index):
    """Merges two indexes by comparing the two files, index.rst and new_index.rst
    section by section and adding pieces which exist in index.rst but are missing
    from new_index.rst . At the end, the one which was added to is maintained as
    index.rst and new_index.rst is deleted.

    Args:
        new_index: path to index file which is being merged from
        old_index: path to existing index file which is being merged into

    Returns:
        none
    True
    """
    old_lines, old_section_limits = find_index_section_limits(old_index)
    new_lines, new_section_limits = find_index_section_limits(new_index)

    for start, end in old_section_limits:
        included = False
        caption = old_lines[start+1]
        for new_start, new_end in new_section_limits:
            if new_lines[new_start+1] == caption:
                new_section = merge_index_sections(new_lines[new_start:new_end], old_lines[start:end])
                del new_lines[new_start:new_end]
                i = new_start
                for line in new_section:
                    new_lines.insert(i, line)
                    i += 1
                included = True
        if not included:
            i = new_end
            new_lines.insert(i, "\n")
            for line in old_lines[start:end]:
                i += 1
                new_lines.insert(i, line)

    old_index_file = open(old_index, "w+")
    old_index_file.write("".join(new_lines))
    old_index_file.close()

    os.remove(new_index)

def find_treatment_section_limits(filename, section_delimiter=".. _heading"):
    """Helper function for merge_treatment_processes which loops through the
    file and marks the beginning and end of each section.

    Args:
        filename: path to file to be modified
        section_delimiter: string which marks the separation between sections
            Default: '.. _heading'

    Returns:
        lines: list of strings of each line in the file
        section_limits: list of the form [[start1, end1], [start2, end2]]
            which marks the separation between sections
    """
    section_limits = []
    start = 0
    file = open(filename, "r+")
    lines = file.readlines()
    file.close()

    for i, line in enumerate(lines):
        if section_delimiter in line:
            end = i - 1
            section_limits.append([start, end])
            start = i

    section_limits.append([start,len(lines)])

    return lines, section_limits

def merge_treatment_processes(new_processes, old_processes):
    """Merges two treatment process descriptions by comparing the two files
    section by section and adding pieces which exist in new_processes but are missing
    from old_processes.

    Args:
        new_processes: path to treatment process file which is being merged from
        old_processes: path to existing treatment process file which is being merged into

    Returns:
        none
    """
    old_lines, old_section_limits = find_treatment_section_limits(old_processes)
    new_lines, new_section_limits = find_treatment_section_limits(new_processes)

    for start, end in new_section_limits:
        included = False
        heading = new_lines[start]
        for old_start, old_end in old_section_limits:
            if old_lines[old_start] == heading:
                included = True
        if not included:
            i = old_end
            old_lines.insert(i, "\n")
            for line in new_lines[start:end]:
                i += 1
                old_lines.insert(i, line)

    old_file = open(old_processes, "w+")
    old_file.write("".join(old_lines))
    old_file.close()

def parse_variables_from_map(unparsed, default_key="", for_docs=True):
    """Helper function for parse_attributes which loops through an unparsed map
    that matched one of the desired fields

    Args:
        unparsed: portion of deserialized JSON which has yet to be parsed
        default_key: key for the field. Used to detect special entries like index
        for_docs: True if parsing variables for AIDE documentation,
            False otherwise (e.g. validation)

    Returns:
        parsed_variables: dictionary of parsed variables
        templates: list of templates to move from doc_files and render in the
            design specs.
        processes: list of unit processes in the given Onshape model
    """
    parsed_variables = {}
    value = None
    templates = []
    processes = []

    if default_key == "template":
        if for_docs:
            copy_to_docs(unparsed)
        templates.append(unparsed)
        return parsed_variables, templates, processes
    elif default_key == "index":
        if unparsed != "" and unparsed is not None and for_docs:
                if os.path.exists('index.rst'):
                    copy_to_docs(unparsed, 'new_index.rst')
                    merge_indexes('new_index.rst', 'index.rst')
                else:
                    copy_to_docs(unparsed, 'index.rst')
        return parsed_variables, templates, processes
    elif default_key == "process":
        processes.append(unparsed)
        if unparsed != "" and unparsed is not None and for_docs:
            file = "Introduction/Treatment_Process.rst"
            file_path = "../../../doc_files/Introduction/Treatment_Process_" + unparsed + ".rst"
            if os.path.exists(file):
                merge_treatment_processes(file_path, file)
            else:
                try:
                    copyfile(file_path, file)
                except IOError as io_err:
                    os.makedirs(os.path.dirname(file))
                    copyfile(file_path, file)
        return parsed_variables, templates, processes

    if isinstance(unparsed, list):
        for to_parse in unparsed:
            if is_fs_type(to_parse, "BTFSValueMapEntry"):
                key = to_parse[msg_str][key_str][msg_str][val_str]
                candidate_message = to_parse[msg_str][val_str]
                if is_fs_type(candidate_message, "BTFSValueMap"):
                    value, template = parse_variables_from_map(candidate_message[msg_str][val_str])
                    templates.extend(template)
                elif is_fs_type(candidate_message,  "BTFSValueArray"):
                    value = parse_variables_from_list(candidate_message[msg_str][val_str],
                                                      for_docs)
                elif is_fs_type(candidate_message, "BTFSValueWithUnits"):
                    value = parse_quantity(candidate_message[msg_str], for_docs)
                elif is_fs_type(candidate_message, ["BTFSValueNumber", "BTFSValueString"]):
                    value = candidate_message[msg_str][val_str]
                parsed_variables[key] = value
    else:
        parsed_variables[default_key] = unparsed

    return parsed_variables, templates, processes

def parse_attributes(attributes, fields, for_docs=True, type_tag="Documenter"):
    """Helper function for get_parsed_measurements which loops through the
    atributes, parsing only the specified fields.

    Args:
        attributes: deserialized JSON object returned by Onshape link
        fields: fields which we are interested in parsing, e.g. 'variables' or 'index'
        for_docs: True if parsing variables for AIDE documentation,
            False otherwise (e.g. validation)
        type_tag: type from Onshape of the configuration we are parsing for
            Default: 'Documenter'

    Returns:
        measurements: dictionary of parsed variables
        templates: list of templates to move from doc_files and render in the
            design specs.
        processes: list of unit processes in the given Onshape model
    """
    measurements = {}
    templates = []
    processes = []

    for attr in attributes:
        if is_fs_type(attr, "BTFSValueMap"):
            if attr[msg_str]["typeTag"] == type_tag:
                for attr2 in attr[msg_str][val_str]:
                    docs = attr2[msg_str][val_str][msg_str][val_str]
                    for doc in docs:
                        for unparsed in doc[msg_str][val_str]:
                            if is_fs_type(unparsed, "BTFSValueMapEntry"):
                                key = unparsed[msg_str][key_str][msg_str][val_str]
                                for field in fields:
                                    if key == field:
                                        new_measure, new_templates, new_processes = parse_variables_from_map(
                                            unparsed[msg_str][val_str][msg_str][val_str],
                                            key,
                                            for_docs
                                        )
                                        measurements.update(new_measure)
                                        templates.extend(new_templates)
                                        processes.extend(new_processes)

    for i in range(len(templates)):
        new_template = './' + os.path.basename(os.path.dirname(templates[i])) + \
                       '/' + os.path.basename(templates[i])
        templates[i] = new_template

    return measurements, templates, processes

def get_parsed_measurements(link,
                            fields=["variables", "template", "index", "process"],
                            for_docs=True):
    """Parses the output of the Onshape Documenter feature found in the Onshape
    document at the given url.

    Args:
        link: URL of Onshape document
        fields: names of fields to search for in the Onshape JSON object
        for_docs: True if parsing variables for AIDE documentation,
            False otherwise (e.g. validation)

    Returns:
        measurements: dictionary of parsed variables
        templates: list of templates to move from doc_files and render in the
            design specs.
        processes: list of unit processes in the given Onshape model
    """
    script = r"""
        function (context is Context, queries is map)
        {
            return getAttributes(context, {
                "entities" : qEverything(),
            });
        }
        """

    client = Client(
        configuration = {
            "base_url": "https://cad.onshape.com",
            "access_key": "ekAHCj04TtODlvlI9yWj2bjB",
            "secret_key": "sS11vEOD5CavkLVcZshLBgfBlB5aBvnpz6v3oEvC0bN0zxhW"
        }
    )

    element = OnshapeElement(link)

    script_call = BTFeatureScriptEvalCall2377(script=script)
    response = client.part_studios_api.eval_feature_script(
        element.did,
        element.wvm,
        element.wvmid,
        element.eid,
        bt_feature_script_eval_call_2377=script_call,
        _preload_content=False,
    )

    attributes = json.loads(response.data.decode("utf-8"))["result"][msg_str][val_str]

    measurements, templates, processes = parse_attributes(attributes, fields, for_docs)

    return measurements, templates, processes

# from https://stackoverflow.com/questions/5914627/prepend-line-to-beginning-of-a-file
def line_prepender(filename, line):
    """Prepends a file with the given line.

    Args:
        filename: path to file to be modified
        line: string of text to prepend to the file

    Returns:
        none
    """
    with open(filename, 'r+') as f:
        content = f.read()
        f.seek(0, 0)
        f.write(line.rstrip('\r\n') + '\n' + content)

def make_replace_list(parsed_dict, filename, var_attachment=''):
    """Adds the dictionary of variables which have been parsed to the top of the
    given file.

    Args:
        parsed_dict: dictionary of variables parsed from Onshape document
        filename: path to file to be modified
        var_attachment: string to prepend to all variables, e.g. "LFOM"
            Default: ''

    Returns:
        none
    """
    prefix = '.. |'
    suffix = '| replace:: '

    for var in parsed_dict:
        if type(parsed_dict[var]) == dict:
            make_replace_list(parsed_dict[var], filename, var_attachment + var + "_")
        else:
            line = prefix + var_attachment + str(var) + suffix + str(parsed_dict[var])
            line_prepender(filename, line)



"""The chemical dose controller (CDC) of an AguaClara plant uses the linear
relation between flow rate and entrance tank water level (set by the LFOM) to
dose the correct amount of coagulant and chlorine into the entrance tank.

Example:
    >>> from aguaclara.design.cdc import *
    >>> cdc = CDC(q = 20 * u.L/u.s, coag_type = 'pacl')
    >>> cdc.coag_stock_vol
    <Quantity(2500.0, 'liter')>
"""
import aguaclara.core.physchem as pc
import aguaclara.core.utility as ut
from aguaclara.core.units import u
from aguaclara.design.component import Component
import warnings

import numpy as np

class CDC(Component):
    """Design an AguaClara plant's chemical dose controller.

    Design Inputs:
        - ``q (float * u.L / u.s)``: Flow rate (required)
    """
    def __init__(self, **kwargs):
        self.hl = 20 * u.cm
        self.coag_type='pacl'
        if self.coag_type.lower() not in ['pacl', 'alum']:
            raise ValueError('coag_type must be either PACl or Alum.')

        self.coag_dose_conc_max=100 * u.mg / u.L
        self.coag_stock_conc=150 * u.g / u.L
        self.coag_stock_conc_est=150 * u.g / u.L # Deprecated since January 2021
        self.coag_stock_min_est_time=1 * u.day
        self.chem_tank_vol_supplier=[208.198, 450, 600, 750, 1100, 2500] * u.L
        self.chem_tank_dimensions_supplier=[
            [0.571, 0.851],
            [0.85, 0.99],
            [0.96, 1.10],
            [1.10, 1.02],
            [1.10, 1.39],
            [1.55, 1.65]
        ] * u.m
        self.train_n=1
        self.coag_sack_mass = 25 * u.kg
        self.coag_tube_id = 0.125 * u.inch
        self.error_ratio=0.1
        self.tube_k = 2

        super().__init__(**kwargs)

    def alum_nu(self, coag_conc):
        """
        Return the dynamic viscosity of water at a given temperature.

        If given units, the function will automatically convert to Kelvin.
        If not given units, the function will assume Kelvin.
        This function assumes that the temperature dependence can be explained
        based on the effect on water and that there is no confounding effect from
        the coagulant.
        """
        alum_nu = \
            (1 + (4.255 * 10 ** -6) * coag_conc.magnitude ** 2.289) * \
            pc.viscosity_kinematic_water(self.temp)
        return alum_nu
    
    def _alum_nu(self, coag_conc):
        """
        .. deprecated::
            `_alum_nu` is deprecated; use `alum_nu` instead.
        """
        # Deprecated since January 2021
        warnings.warn('_alum_nu is deprecated; use alum_nu instead.',
                  UserWarning)
    
        return self.alum_nu(coag_conc)
        
    def pacl_nu(self, coag_conc):
        """Return the dynamic viscosity of water at a given temperature.

        If given units, the function will automatically convert to Kelvin.
        If not given units, the function will assume Kelvin.
        This function assumes that the temperature dependence can be explained
        based on the effect on water and that there is no confounding effect 
        from the coagulant.
        """
        pacl_nu = \
            (1 + (2.383 * 10 ** -5) * (coag_conc).magnitude ** 1.893) * \
            pc.viscosity_kinematic_water(self.temp)
        return pacl_nu

    def _pacl_nu(self, coag_conc):
        """
        .. deprecated::
            `_pacl_nu` is deprecated; use `pacl_nu` instead.
        """
        # Deprecated since January 2021
        warnings.warn('_pacl_nu is deprecated; use pacl_nu instead.',
                  UserWarning)
    
        return self.pacl_nu(coag_conc)
        
    def _coag_nu(self, coag_conc, coag_type):
        """
        .. deprecated::
            `_coag_nu` is deprecated; use `coag_nu` instead.
        """
        # Deprecated since January 2021
        warnings.warn('_coag_nu is deprecated; use coag_nu instead.',
                  UserWarning)
    
        return self.coag_nu(coag_conc, coag_type)
       
    def coag_nu(self, coag_conc, coag_type):
        """Return the dynamic viscosity of water at a given temperature.

        If given units, the function will automatically convert to Kelvin.
        If not given units, the function will assume Kelvin.
        """
        if coag_type.lower() == 'alum':
            coag_nu = self.alum_nu(coag_conc)
        elif coag_type.lower() == 'pacl':
            coag_nu = self.pacl_nu(coag_conc)
        return coag_nu

    @property
    def coag_q_max_est(self):
        """The estimated maximum permissible flow rate of the coagulant stock,
        based on a whole number of sacks of coagulant used.

        .. deprecated::
            `coag_q_max_est` is deprecated; use `coag_q_max` instead, which is
            based on the exact user-defined coagulant stock concentration, 
            `coag_stock_conc`.
        """
        # Deprecated since January 2021
        warnings.warn('coag_q_max_est is deprecated; use coag_q_max instead,\
            which is based on the exact user-defined coagulant stock \
            concentration, coag_stock_conc.', UserWarning)
        
        coag_q_max_est = self.q * self.coag_dose_conc_max / \
            self.coag_stock_conc_est
        return coag_q_max_est

    @property
    def coag_q_max(self):
        """The maximum permissible flow rate of the coagulant stock."""
        coag_q_max = self.q * self.coag_dose_conc_max / self.coag_stock_conc
        return coag_q_max.to(u.L / u.s)

    @property
    def coag_stock_vol(self):
        """The volume of the coagulant stock tank, based on available sizes
        from the supplier.
        """
        coag_stock_vol = ut.ceil_nearest(
                self.coag_stock_min_est_time * self.train_n *
                    self.coag_q_max,
                self.chem_tank_vol_supplier
            )
        return coag_stock_vol

    @property
    def coag_sack_n(self):
        """The number of sacks of coagulant needed to make the coagulant stock,
        rounded to the nearest whole number.
        """
        coag_sack_n = round(
                (self.coag_stock_vol * self.coag_stock_conc /
                self.coag_sack_mass).to_base_units()
            )
        return coag_sack_n

    # Commented out January 2021
    # @property
    # def coag_stock_conc(self):
    #     """The concentration of the coagulant stock."""
    #     coag_stock_conc = self.coag_sack_n * self.coag_sack_mass / \
    #         self.coag_stock_vol
    #     return coag_stock_conc

    @property
    def coag_stock_time_min(self):
        """The minimum amount of time that the coagulant stock will last."""
        return self.coag_stock_vol / (self.train_n * self.coag_q_max)

    @property
    def coag_stock_nu(self):
        """The kinematic viscosity of the coagulant stock."""
        return self.coag_nu(self.coag_stock_conc, self.coag_type)
#==============================================================================
# Small-diameter Tube Design
#==============================================================================
    @property
    def _coag_tube_q_max(self):
        """The maximum permissible flow through a coagulant tube."""
        coag_tube_q_max = ((np.pi * self.coag_tube_id ** 2)/4) * \
            np.sqrt((2 * self.error_ratio * self.hl * u.gravity)/self.tube_k)
        return coag_tube_q_max.to(u.L / u.s)

    @property
    def coag_tubes_active_n(self):
        """The number of coagulant tubes in use."""
        coag_tubes_active_n = \
            np.ceil((self.coag_q_max / self._coag_tube_q_max).to_base_units())
        return coag_tubes_active_n

    @property
    def coag_tubes_n(self):
        """The number of coagulant tubes in use, plus a spare tube for 
        maintenance.
        """
        coag_tubes_n = self.coag_tubes_active_n + 1
        return coag_tubes_n

    @property
    def coag_tube_operating_q_max(self):
        """The maximum flow through a coagulant tube during actual operation."""
        coag_tube_operating_q_max = self.coag_q_max / self.coag_tubes_active_n
        return coag_tube_operating_q_max

    @property
    def coag_tube_l(self):
        """The length of a coagulant tube."""
        coag_tube_l = (
                self.hl * u.gravity * np.pi * self.coag_tube_id ** 4 /
                (128 * self.coag_stock_nu * self.coag_tube_operating_q_max)
            ) - (
                8 * self.coag_tube_operating_q_max * self.tube_k /
                (128 * np.pi * self.coag_stock_nu)
            )
        return coag_tube_l.to_base_units()

    @property
    def coag_tank_r(self):
        """The radius of the coagulant stock tank, based on available sizes
        from the supplier."""
        index = np.argmax(self.chem_tank_vol_supplier == self.coag_stock_vol)
        coag_tank_r = self.chem_tank_dimensions_supplier[index][0] / 2
        return coag_tank_r

    @property
    def coag_tank_h(self):
        """The height of the coagulant stock tank, based on available sizes
        from the supplier."""
        index = np.argmax(self.chem_tank_vol_supplier == self.coag_stock_vol)
        coag_tank_h = self.chem_tank_dimensions_supplier[index][1]
        return coag_tank_h

    def _DiamTubeAvail(self, en_tube_series = True):
        if en_tube_series:
            return 1*u.mm
        else:
            return (1/16)*u.inch



from aguaclara.core.units import u
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from datetime import datetime, timedelta
import os
from pathlib import Path


def column_of_data(path, start, column, end=None, units=""):
    """This function extracts a column of data from a ProCoDA data file.

    Note: Column 0 is time. The first data column is column 1.

    :param path: The file path of the ProCoDA data file
    :type path: string
    :param start: Index of first row of data to extract, inclusive
    :type start: int
    :param end: Index of last row of data to extract until, exclusive. Defaults to extracting all rows.
    :type end: int, optional
    :param column: Index or label of the column that you want to extract
    :type column: int or string
    :param units: The units you want to apply to the data, e.g. 'mg/L'. Defaults to "" (dimensionless).
    :type units: string, optional

    :return: The column of data
    :rtype: numpy.ndarray in units of [units]

    :Examples:

    .. code-block:: python

        data = column_of_data("Reactor_data.txt", 0, 1, -1, "mg/L")
    """
    df = pd.read_csv(path, delimiter='\t')

    if isinstance(column, int):
        data = df.iloc[start:end, column]
    else:
        data = df.iloc[start:end][column]
    num_data = data[pd.to_numeric(data, errors='coerce').notnull()]

    return np.array(num_data) * u(units)


def column_of_time(path, start, end=None, units="day"):
    """This function extracts the column of times as elasped times from a ProCoDA data file.

    :param path: The file path of the ProCoDA data file.
    :type path: string
    :param start: Index of first row of data to extract from the data file
    :type start: int
    :param end: Index of last row of data to extract from the data. Defaults to last row
    :type end: int
    :param units: The return type units, which defaults to day.
    :type units: string, optional

    :return: Experimental times starting at 0
    :rtype: numpy.ndarray in units of days or hours, specified with units parameter

    :Examples:

    .. code-block:: python

        time = column_of_time("Reactor_data.txt", 0)
    """
    df = pd.read_csv(path, delimiter='\t')

    start_time = pd.to_numeric(df.iloc[start, 0])
    day_times = df.iloc[start:end, 0]
    is_numeric = pd.to_numeric(day_times, errors='coerce').notnull()
    num_day_times = pd.to_numeric(day_times[is_numeric])
    elapsed_times = num_day_times - start_time

    return (np.array(elapsed_times) * u.day).to(u(units))


def plot_columns(path, columns, x_axis=None):
    """Plot columns of data, located by labels, in the given data file.

    :param path: The file path of the ProCoDA data file
    :type path: string
    :param columns: A single column label or list of column labels
    :type columns: string or string list
    :param x_axis: The label of the x-axis column (defaults to None)
    :type x_axis: string, optional

    :return: A list of Line2D objects representing the plotted data
    :rtype: matplotlib.lines.Line2D list
    """
    df = pd.read_csv(path, delimiter='\t')
    df = remove_notes(df)

    if isinstance(columns, str):
        y = pd.to_numeric(df.loc[:, columns])
        if x_axis is None:
            plt.plot(y)
        else:
            x = pd.to_numeric(df.loc[:, x_axis])
            plt.plot(x, y)

    elif isinstance(columns, list):
        for c in columns:
            y = pd.to_numeric(df.loc[:, c])
            if x_axis is None:
                plt.plot(y)
            else:
                x = pd.to_numeric(df.loc[:, x_axis])
                plt.plot(x, y)
    else:
        raise ValueError('columns must be a string or list of strings')


def iplot_columns(path, columns, x_axis=None):
    """Plot columns of data, located by indexes, in the given data file.

    :param path: The file path of the ProCoDA data file
    :type path: string
    :param columns: A single column index or list of column indexes
    :type columns: int or int list
    :param x_axis: The index of the x-axis column (defaults to None)
    :type x_axis: int, optional
    :param sep: The separator or delimiter, of the data file. Use ',' for CSV's, '\t' for TSV's.
    :type sep: string

    :return: a list of Line2D objects representing the plotted data
    :rtype: matplotlib.lines.Line2D list
    """
    df = pd.read_csv(path, delimiter='\t')
    df = remove_notes(df)

    if isinstance(columns, int):
        y = pd.to_numeric(df.iloc[:, columns])
        if x_axis is None:
            plt.plot(y)
        else:
            x = pd.to_numeric(df.iloc[:, x_axis])
            plt.plot(x, y)

    elif isinstance(columns, list):
        for c in columns:
            y = pd.to_numeric(df.iloc[:, c])
            if x_axis is None:
                plt.plot(y)
            else:
                x = pd.to_numeric(df.iloc[:, x_axis])
                plt.plot(x, y)
    else:
        raise ValueError('columns must be an int or a list of ints')


def notes(path):
    """This function extracts any experimental notes from a ProCoDA data file.
    Use this to identify the section of the data file that you want to extract.

    :param path: The file path of the ProCoDA data file.
    :type path: string

    :return: The rows of the data file that contain text notes inserted during the experiment.
    :rtype: pandas.Dataframe
    """
    df = pd.read_csv(path, delimiter='\t')
    return df[pd.to_numeric(df.iloc[:, 0], errors='coerce').isnull()]


def remove_notes(data):
    """Omit notes from a DataFrame object, where notes are identified as rows
    with non-numerical entries in the first column.

    :param data: DataFrame object to remove notes from
    :type data: Pandas.DataFrame

    :return: DataFrame object with no notes
    :rtype: Pandas.DataFrame
    """
    return data[pd.to_numeric(data.iloc[:, 0], errors='coerce').notnull()]


def get_data_by_time(path, columns, dates, start_time='00:00', end_time='23:59',
                     extension='.tsv', units='', elapsed=False):
    """Extract columns of data over one or more ProCoDA data files based on date
    and time. Valid only for files whose names are automatically generated by
    date, i.e. of the form "datalog_M-D-YYYY".

    Note: Column 0 is time. The first data column is column 1. Results for the
    time column are adjusted for multi-day experiments.

    :param path: The path to the folder containing the ProCoDA data file(s)
    :type path: string
    :param columns: A single column index or a list of column indexes
    :type columns: int or int list
    :param dates: A single date or list of dates, formatted "M-D-YYYY"
    :type dates: string or string list
    :param start_time: Starting time of data to extract, formatted 'HH:MM' (24-hour time)
    :type start_time: string, optional
    :param end_time: Ending time of data to extract, formatted 'HH:MM' (24-hour time)
    :type end_time: string, optional
    :param extension: File extension of the data file(s). Defaults to '.tsv'
    :type extension: string, optional
    :param units: A single unit or list of units to apply to each column, e.g. 'mg/L' or ['hr', 'mg/L']. Defaults to '' (dimensionless).
    :type units: string, optional
    :param elapsed: If true, results for the time column are given in elapsed time
    :type elapsed: boolean

    :return: the single column of data or a list of the columns of data (in the order of the indexes given in the columns variable)
    :rtype: 1D or 2D float list

    :Examples:

    .. code-block:: python

        data = get_data_by_time(path='/Users/.../ProCoDA Data/', columns=4, dates=['6-14-2018', '6-15-2018'], start_time='12:20', end_time='10:50')
        data = get_data_by_time(path='/Users/.../ProCoDA Data/', columns=[0,4], dates='6-14-2018', start_time='12:20', end_time='23:59')
        data = get_data_by_time(path='/Users/.../ProCoDA Data/', columns=[0,3,4], dates='6-14-2018')
    """
    # the file path url is not acceptable (ie contains 'github.com')
    if 'github.com' in path:
        path = path.replace('github.com', 'raw.githubusercontent.com')
        path = path.replace('blob/', '')
        path = path.replace('tree/', '')

    data = data_from_dates(path, dates, extension) # combine data from each date

    first_time_column = pd.to_numeric(data[0].iloc[:, 0])
    start = max(day_fraction(start_time), first_time_column[0])
    start_idx = (first_time_column >= start).idxmax()
    end_idx = (pd.to_numeric(data[-1].iloc[:, 0]) >=
               day_fraction(end_time)).idxmax() + 1

    if isinstance(columns, int):
        if columns == 0 and elapsed:
            col = column_start_to_end(data, columns, start_idx, end_idx)
            result = list(np.subtract(col, start))*u(units)
        else:
            result = column_start_to_end(data, columns, start_idx, end_idx)*u(units)
    else: # columns is a list
        if units == '':
            units = ['']*len(columns)
        result = []
        i = 0
        for c in columns:
            if c == 0 and elapsed:
                col = column_start_to_end(data, c, start_idx, end_idx)
                result.append(list(np.subtract(col, start))*u(units[i]))
            else:
                result.append(column_start_to_end(data, c, start_idx, end_idx)*u(units[i]))
            i += 1

    return result


def day_fraction(time):
    """Convert a 24-hour time to a fraction of a day. For example, midnight
    corresponds to 0.0, and noon to 0.5.

    :param time: Time in the form of 'HH:MM' (24-hour time)
    :type time: string

    :return: A day fraction
    :rtype: float

    :Examples:

    .. code-block:: python

        day_fraction("18:30")
    """
    hour = int(time.split(":")[0])
    minute = int(time.split(":")[1])
    return hour/24 + minute/1440


def data_from_dates(path, dates, extension):
    """Return a list of DataFrames representing the ProCoDA data files stored in
    the given path and recorded on the given dates.

    :param path: The path to the folder containing the ProCoDA data file(s)
    :type path: string
    :param dates: A single date or list of dates for which data was recorded, formatted "M-D-YYYY"
    :type dates: string or string list
    :param extension: File extension of the data file(s)
    :type extension: string, optional

    :return: a list of DataFrames representing the ProCoDA data files recorded on the given dates
    :rtype: pandas.DataFrame list
    """
    if not isinstance(dates, list):
        dates = [dates]

    data = []
    for d in dates:
        filepath = os.path.join(path, 'datalog_' + d + extension)
        data.append(remove_notes(pd.read_csv(filepath, delimiter='\t')))

    return data


def column_start_to_end(data, column, start_idx, end_idx):
    """Return a list of numeric data entries in the given column from the starting
    index to the ending index. This can list can be compiled over one or more
    DataFrames.

    :param data: a list of DataFrames to extract one column from
    :type data: Pandas.DataFrame list
    :param column: a column index
    :type column: int
    :param start_idx: the index of the starting row of the first DataFrame
    :type start_idx: int
    :param end_idx: the index of the ending row of the last DataFrame, excluding this row
    :type end_idx: int

    :return: a list of data from the given column
    :rtype: float list
    """
    if len(data) == 1:
        result = list(pd.to_numeric(data[0].iloc[start_idx:end_idx, column]))
    else:
        result = list(pd.to_numeric(data[0].iloc[start_idx:, column]))
        for i in range(1, len(data)-1):
            data[i].iloc[0, 0] = 0
            result += list(pd.to_numeric(data[i].iloc[:, column]) +
                      (i if column == 0 else 0))
                      # assuming DataFrames are for consecutive days, add number of
                      # DataFrame if dealing with the time column (column 0)
        data[-1].iloc[0, 0] = 0
        result += list(pd.to_numeric(data[-1].iloc[:end_idx, column]) +
                  (len(data)-1 if column == 0 else 0))

    return result


def get_data_by_state(path, dates, state, column, extension=".tsv"):
    """Reads a ProCoDA file and extracts the time and data column for each
    iteration of the given state.

    Note: column 0 is time, the first data column is column 1. Results for the
    time column are given in elasped time.

    :param path: The path to the folder containing the ProCoDA data file(s), defaults to the current directory
    :type path: string
    :param dates: A single date or list of dates for which data was recorded, formatted "M-D-YYYY"
    :type dates: string or string list
    :param state: The state ID number for which data should be extracted
    :type state: int
    :param column: The integer index of the column that you want to extract OR the header of the column that you want to extract
    :type column: int or string
    :param extension: File extension of the data file(s). Defaults to '.tsv'
    :type extension: string, optional

    :return: A list of lists of the time and data columns extracted for each iteration of the state. For example, if "data" is the output, data[i][:,0] gives the time column and data[i][:,1] gives the data column for the ith iteration of the given state and column. data[i][0] would give the first [time, data] pair.
    :type: 3D float list

    :Examples:

    .. code-block:: python

        data = get_data_by_state(path='/Users/.../ProCoDA Data/', dates=["6-19-2013", "6-20-2013"], state=1, column=28)
    """
    # the file path url is not acceptable (ie contains 'github.com')
    if 'github.com' in path:
        path = path.replace('github.com', 'raw.githubusercontent.com')
        path = path.replace('blob/', '')
        path = path.replace('tree/', '')

    data_agg = []
    day = 0
    first_day = True
    overnight = False
    if path[-1] != '/':
        path += '/'

    if not isinstance(dates, list):
        dates = [dates]

    for d in dates:
        state_file = path + "statelog_" + d + extension
        data_file = path + "datalog_" + d + extension

        states = pd.read_csv(state_file, delimiter='\t')
        data = pd.read_csv(data_file, delimiter='\t')

        states = np.array(states)
        data = np.array(data)

        # get the start and end times for the state
        state_start_idx = states[:, 1] == state
        state_start = states[state_start_idx, 0]
        state_end_idx = np.append([False], state_start_idx[0:-1])
        state_end = states[state_end_idx, 0]

        if overnight:
            state_start = np.insert(state_start, 0, 0)
            state_end = np.insert(state_end, 0, states[0, 0])

        if state_start_idx[-1]:
            state_end = np.append(state_end, data[-1, 0])

        # get the corresponding indices in the data array
        data_start = []
        data_end = []
        for i in range(np.size(state_start)):
            data_start.append((data[:, 0] > state_start[i]).argmax())
            data_end.append((data[:, 0] > state_end[i]).argmax()-1)
        if np.size(data_end) < np.size(data_start):
            data_end = np.append(data_end, -1)

        if first_day:
            start_time = data[0, 0]

        for i in range(np.size(data_start)):
            t = data[data_start[i]:data_end[i], 0] + day - start_time
            if isinstance(column, int):
                c = data[data_start[i]:data_end[i], column]
            else:
                c = data[column][data_start[i]:data_end[i]]
            if overnight and i == 0:
                data_agg = np.insert(data_agg[-1], np.size(data_agg[-1][:, 0]),
                                     np.vstack((t, c)).T)
            else:
                data_agg.append(np.vstack((t, c)).T)

        day += 1
        if first_day:
            first_day = False
        if state_start_idx[-1]:
            overnight = True

    return data_agg


def read_state(dates, state, column, units="", path="", extension=".tsv"):
    """Reads a ProCoDA file and outputs the data column and time vector for
    each iteration of the given state.

    Note: Column 0 is time. The first data column is column 1.

    :param dates: A single date or list of dates for which data was recorded, formatted "M-D-YYYY"
    :type dates: string or string list
    :param state: The state ID number for which data should be extracted
    :type state: int
    :param column: Index of the column that you want to extract OR header of the column that you want to extract
    :type column: int or string
    :param units: The units you want to apply to the data, e.g. 'mg/L'. Defaults to "" (dimensionless)
    :type units: string, optional
    :param path: The file path of the ProCoDA data file.
    :type path: string
    :param extension: The file extension of the tab delimited file. Defaults to ".tsv"
    :type extension: string, optional

    :return: time (numpy.ndarray) - Times corresponding to the data (with units)
    :return: data (numpy.ndarray) - Data in the given column during the given state with units

    :Examples:

    .. code-block:: python

        time, data = read_state(["6-19-2013", "6-20-2013"], 1, 28, "mL/s")
    """
    data_agg = get_data_by_state(path, dates, state, column, extension)
    data_agg = np.vstack(data_agg)
    if units != "":
        return data_agg[:, 0]*u.day, data_agg[:, 1]*u(units)
    else:
        return data_agg[:, 0]*u.day, data_agg[:, 1]


def average_state(dates, state, column, units="", path="", extension=".tsv"):
    """Outputs the average value of the data for each instance of a state in
    the given ProCoDA files

    Note: Column 0 is time. The first data column is column 1.

    :param dates: A single date or list of dates for which data was recorded, formatted "M-D-YYYY"
    :type dates: string or string list
    :param state: The state ID number for which data should be extracted
    :type state: int
    :param column: Index of the column that you want to extract OR header of the column that you want to extract
    :type column: int or string
    :param units: The units you want to apply to the data, e.g. 'mg/L'. Defaults to "" (dimensionless)
    :type units: string, optional
    :param path: The file path of the ProCoDA data file.
    :type path: string
    :param extension: The file extension of the tab delimited file. Defaults to ".tsv"
    :type extension: string, optional

    :return: A list of averages for each instance of the given state
    :rtype: float list

    :Examples:

    .. code-block:: python

        data_avgs = average_state(["6-19-2013", "6-20-2013"], 1, 28, "mL/s")

    """
    data_agg = get_data_by_state(path, dates, state, column, extension)

    averages = np.zeros(np.size(data_agg))
    for i in range(np.size(data_agg)):
        averages[i] = np.average(data_agg[i][:,1])

    if units != "":
        return averages*u(units)
    else:
        return averages


def perform_function_on_state(func, dates, state, column, units="", path="", extension=".tsv"):
    """Performs the function given on each state of the data for the given state
    in the given column and outputs the result for each instance of the state

    Note: Column 0 is time. The first data column is column 1.

    :param func: A function that will be applied to data from each instance of the state
    :type func: function
    :param dates: A single date or list of dates for which data was recorded, formatted "M-D-YYYY"
    :type dates: string or string list
    :param state: The state ID number for which data should be extracted
    :type state: int
    :param column: Index of the column that you want to extract OR header of the column that you want to extract
    :type column: int or string
    :param units: The units you want to apply to the data, e.g. 'mg/L'. Defaults to "" (dimensionless)
    :type units: string, optional
    :param path: The file path of the ProCoDA data file.
    :type path: string
    :param extension: The file extension of the tab delimited file. Defaults to ".tsv".
    :type extension: string, optional

    :requires: func takes in a list of data with units and outputs the correct units

    :return: The outputs of the given function for each instance of the given state
    :type: list

    :Examples:

    .. code-block:: python

        def avg_with_units(lst):
        num = np.size(lst)
        acc = 0
        for i in lst:
            acc = i + acc

        return acc / num

        data_avgs = perform_function_on_state(avg_with_units, ["6-19-2013", "6-20-2013"], 1, 28, "mL/s")
    """
    data_agg = get_data_by_state(path, dates, state, column, extension)

    output = np.zeros(np.size(data_agg))
    for i in range(np.size(data_agg)):
        if units != "":
            output[i] = func(data_agg[i][:,1]*u(units)).magnitude
        else:
            output[i] = func(data_agg[i][:,1])

    if units != "":
        return output*func(data_agg[i]*u(units)).units
    else:
        return output


def read_state_with_metafile(func, state, column, path, metaids=[],
                             extension=".tsv", units=""):
    """Takes in a ProCoDA meta file and performs a function for all data of a
    certain state in each of the experiments (denoted by file paths in then
    metafile)

    Note: Column 0 is time. The first data column is column 1.

    :param func: A function that will be applied to data from each instance of the state
    :type func: function
    :param state: The state ID number for which data should be extracted
    :type state: int
    :param column: Index of the column that you want to extract OR header of the column that you want to extract
    :type column: int or string
    :param path: The file path of the ProCoDA data file (must be tab-delimited)
    :type path: string
    :param metaids: a list of the experiment IDs you'd like to analyze from the metafile
    :type metaids: string list, optional
    :param extension: The file extension of the tab delimited file. Defaults to ".tsv"
    :type extension: string, optional
    :param units: The units you want to apply to the data, e.g. 'mg/L'. Defaults to "" (dimensionless)
    :type units: string, optional

    :return: ids (string list) - The list of experiment ids given in the metafile
    :return: outputs (list) - The outputs of the given function for each experiment

    :Examples:

    .. code-block:: python

        def avg_with_units(lst):
            num = np.size(lst)
            acc = 0
            for i in lst:
                acc = i + acc

            return acc / num

        path = "../tests/data/Test Meta File.txt"
        ids, answer = read_state_with_metafile(avg_with_units, 1, 28, path, [], ".tsv", "mg/L")
    """
    outputs = []

    metafile = pd.read_csv(path, delimiter='\t', header=None)
    metafile = np.array(metafile)

    ids = metafile[1:, 0]

    if not isinstance(ids[0], str):
        ids = list(map(str, ids))

    if metaids:
        paths = []
        for i in range(len(ids)):
            if ids[i] in metaids:
                paths.append(metafile[i, 4])
    else:
        paths = metafile[1:, 4]

    basepath = os.path.join(os.path.split(path)[0], metafile[0, 4])

    # use a loop to evaluate each experiment in the metafile
    for i in range(len(paths)):
        # get the range of dates for experiment i
        day1 = metafile[i+1, 1]

        # modify the metafile date so that it works with datetime format
        if not (day1[2] == "-" or day1[2] == "/"):
            day1 = "0" + day1
        if not (day1[5] == "-" or day1[5] == "/"):
            day1 = day1[:3] + "0" + day1[3:]

        if day1[2] == "-":
            dt = datetime.strptime(day1, "%m-%d-%Y")
        else:
            dt = datetime.strptime(day1, "%m/%d/%y")
        duration = metafile[i+1, 3]

        if not isinstance(duration, int):
            duration = int(duration)

        date_list = []
        for j in range(duration):
            curr_day = dt.strftime("%m-%d-%Y")
            if curr_day[3] == "0":
                curr_day = curr_day[:3] + curr_day[4:]
            if curr_day[0] == "0":
                curr_day = curr_day[1:]

            date_list.append(curr_day)

            dt = dt + timedelta(days=1)

        path = str(Path(os.path.join(basepath, paths[i]))) + os.sep
        _, data = read_state(date_list, state, column, units, path, extension)

        outputs.append(func(data))

    return ids, outputs


def write_calculations_to_csv(funcs, states, columns, path, headers, out_name,
                              metaids=[], extension=".tsv"):
    """Writes each output of the given functions on the given states and data
    columns to a new column in the specified output file.

    Note: Column 0 is time. The first data column is column 1.

    :param funcs: A function or list of functions which will be applied in order to the data. If only one function is given it is applied to all the states/columns
    :type funcs: function or function list
    :param states: The state ID numbers for which data should be extracted. List should be in order of calculation or if only one state is given then it will be used for all the calculations
    :type states: string or string list
    :param columns: The index of a column, the header of a column, a list of indexes, OR a list of headers of the column(s) that you want to apply calculations to
    :type columns: int, string, int list, or string list
    :param path: Path to your ProCoDA metafile (must be tab-delimited)
    :type path: string
    :param headers: List of the desired header for each calculation, in order
    :type headers: string list
    :param out_name: Desired name for the output file. Can include a relative path
    :type out_name: string
    :param metaids: A list of the experiment IDs you'd like to analyze from the metafile
    :type metaids: string list, optional
    :param extension: The file extension of the tab delimited file. Defaults to ".tsv"
    :type extension: string, optional

    :requires: funcs, states, columns, and headers are all of the same length if they are lists. Some being lists and some single values are okay.

    :return: out_name.csv (CVS file) - A CSV file with the each column being a new calcuation and each row being a new experiment on which the calcuations were performed
    :return: output (Pandas.DataFrame)- Pandas DataFrame holding the same data that was written to the output file
    """
    if not isinstance(funcs, list):
        funcs = [funcs] * len(headers)

    if not isinstance(states, list):
        states = [states] * len(headers)

    if not isinstance(columns, list):
        columns = [columns] * len(headers)

    data_agg = []
    for i in range(len(headers)):
        ids, data = read_state_with_metafile(funcs[i], states[i], columns[i],
                                             path, metaids, extension)
        data_agg = np.append(data_agg, [data])

    output = pd.DataFrame(data=np.vstack((ids, data_agg)).T,
                          columns=["ID"]+headers)
    output.to_csv(out_name, sep='\t')

    return output


def intersect(x, y1, y2):
    """Returns the intersections of two lines represented by a common set of x coordinates and
    two sets of y coordinates as three numpy arrays: the x coordinates of the intersections,
    the y coordinates of the intersections, and the indexes in x, y1, y2 immediately
    after the intersections.

    :param x: common set of x coordinates for the two lines
    :type x: numpy.ndarray
    :param y1: the y coordinates of the first line
    :type y1: numpy.ndarray
    :param y2: the y coordinates of the second line
    :type y2: numpy.ndarray

    :requires: x have no repeating values and is in ascending order

    :return: x_points-numpy.ndarray of the x coordinates where intersections occur
    :return: y_points-numpy.ndarray of the y coordinates where intersections occur
    :return: crossings-numpy.ndarray of the indexes after the intersections occur
    """
    x_points = np.array([])
    y_points = np.array([])
    crossings = (np.argwhere(np.diff(np.sign(y1-y2)))+1).flatten()

    for c in crossings:
      slope1 = (y1[c] - y1[c-1]) / (x[c] - x[c-1])
      slope2 = (y2[c] - y2[c-1]) / (x[c] - x[c-1])
      b1 = y1[c] - slope1 * x[c]
      b2 = y2[c] - slope2 * x[c]

      x_points = np.append(x_points, (b2-b1)/(slope1-slope2))
      y_points = np.append(y_points, slope1*(b2-b1)/(slope1-slope2) + b1)

    return x_points, y_points, crossings

    



"""The sedimentation channel of an AguaClara water treatment plant ensures that
the sedimentation tanks do not exceed their maximum influent flow. If that were
to happen, the upflow velocity would be too great for flocs to settle out.

Example:
    >>> from aguaclara.design.sed_chan import *
    >>> sed_chan = SedimentationChannel(q = 20 * u.L / u.s, temp = 20 * u.degC)
    >>> round(sed_chan.inlet_w, 5)
    <Quantity(59.52756, 'inch')>
"""
from aguaclara.core.units import u
import aguaclara.core.constants as con
import aguaclara.core.materials as mat
import aguaclara.core.utility as ut
import aguaclara.core.physchem as pc
import aguaclara.core.pipes as pipe
import aguaclara.core.head_loss as hl
import aguaclara.design.human_access as ha
from aguaclara.design.component import Component
from aguaclara.design.pipeline import Pipe

import numpy as np
import math


class SedimentationChannel(Component):
    """Design an AguaClara sedimentation channel.

    The sedimentation channel relies on the number and dimensions of the
    sedimentation tanks in the same plant, but assumed/default values may be
    used to design a sedimentation channel by itself. To design these components
    in tandem, use :class:`aguaclara.design.sed.Sedimentor`.

    Constants:
        - ``SED_TANK_Q_RATIO (float)``: Permissible ratio of influent flow
          between the sedimentation tanks
        - ``PLANT_FREE_BOARD_H (float * u.cm)``: Freeboard height in the plant
        - ``WEIR_FREEBOARD_H (float * u.cm)``: Freeboard height of a channel
          weir
        - ``SED_DEPTH_EST (float * u.m)``: Estimated depth of the sedimentor

    Design Inputs:
        - ``q (float * u.L / u.s)``: Flow rate (recommended, defaults to 20L/s)
        - ``temp (float * u.degC)``: Water temperature (recommended, defaults to
          20°C)
        - ``sed_tank_n (int)``: Number of sedimentation tanks (recommended,
          defaults to 4)
        - ``sed_tank_w_inner (float * u.inch)``: Inner width of the sedimentation
          tank (recommended, defaults to 42 in)
        - ``sed_tank_wall_thickness (float * u.cm)``: Wall thickness of the
          sedimentation tank (recommended, defaults to 15 cm)
        - ``sed_tank_inlet_man_nd (float * u.cm)``: Nominal diameter of the
          sedimentation tank's inlet manifold (recommended, defaults to 60 cm)
        - ``sed_tank_outlet_man_nd (float * u.cm)``: Nominal diameter of the
          sedimentation tank's outlet manifold (recommended, defaults to 60 cm)
        - ``sed_tank_outlet_man_hl (float * u.cm)``: Head loss in the
          sedimentation tank's outlet manifold (recommended, defaults to 4 cm)
        - ``sed_tank_diffuser_hl (float * u.mm)``: Head loss through a
          diffuser in the sedimentation tank (recommended, defaults to 0.09 cm)
        - ``sed_wall_thickness (float * u.cm)``: Wall thickness of the
          sedimentor (recommended, defaults to 15 cm)
        - ``weir_thickness (float * u.cm)``: Weir thickness (optional, defaults
          to 15 cm)
        - ``weir_hl (float * u.cm)``: Head loss over the weir (optional,
          defaults to 5 cm)
        - ``w_min (float * u.cm)``: Minimum width (optional, defaults to 30 cm)
        - ``fitting_s (float * u.cm)``: Fitting spacing (optional, defaults
          to 15 cm)
        - ``inlet_depth_max (float * u.cm)``: Maximum inlet channel depth
          (optional, defaults to 50 cm)
        - ``drain_sdr (int)``: SDR of the drain pipe (optional, defaults to 26)
        - ``outlet_free_h (float * u.cm)``: Permissible increase of water level
          in the outlet channel (optional, defaults to 5 cm)
          ``outlet_weir_depth``:The depth of the outlet weir. (optional, defaults to 5 cm)
        - ``outlet_pipe_sdr (int)``: SDR of the outlet pipe (optional, defaults
          to 41)
        - ``outlet_pipe_hl_max (float * u.cm)``: Maximum head loss through the
          outlet pipe (optional, defaults to 1 cm)
        - ``outlet_pipe_nd_max (float * u.inch)``: Maximum nominal diameter of
          the outlet pipe (optional, defaults to 8 in)
    """
    SED_TANK_Q_RATIO = 0.95
    PLANT_FREEBOARD_H = 5.0 * u.cm
    WEIR_FREEBOARD_H = 2.0 * u.cm
    SED_DEPTH_EST = 2.0 * u.m

    def __init__(self, **kwargs):
        self.sed_tank_n=4
        self.sed_tank_w_inner=42.0 * u.inch
        self.sed_tank_wall_thickness = 15.0 * u.cm
        self.sed_tank_inlet_man_nd = 60.0 * u.cm
        self.sed_tank_outlet_man_nd = 60.0 * u.cm
        self.sed_tank_outlet_man_hl = 4.0 * u.cm
        self.sed_tank_diffuser_hl=0.09 * u.mm
        self.sed_wall_thickness = 15.0 * u.cm

        self.weir_thickness = 15.0 * u.cm
        self.weir_hl = 5 * u.cm
        self.w_min = 30.0 * u.cm
        self.fitting_s = 15. * u.cm
        self.inlet_depth_max = 50 * u.cm
        self.drain_spec = 'sdr26'
        self.outlet_free_h = 5.0 * u.cm
        self.outlet_pipe_spec = 'sdr41'
        self.outlet_pipe_hl_max = 1.0 * u.cm
        self.outlet_pipe_nd_max = 8.0 * u.inch

        self.drain_pipe = Pipe()
        self.outlet_pipe = Pipe()
        self.subcomponents = [self.drain_pipe, self.outlet_pipe]

        super().__init__(**kwargs)
        self._set_drain_pipe()
        self._set_outlet_pipe()
        super().set_subcomponents()

    @property
    def l(self):
        """Length of the sedimentation channel."""
        l = (self.sed_tank_n * self.sed_tank_w_inner) + \
            ((self.sed_tank_n-1) * self.sed_tank_wall_thickness) + \
            self.sed_wall_thickness
        return l.to(u.m)

    @property
    def outlet_weir_hl(self):
        """Head loss over the outlet channel weir."""
        weir_exit_hl = pc.headloss_weir_rect(self.q, self.l)
        return weir_exit_hl

    @property
    def inlet_hl_max(self):
        """Maximum head loss in the inlet channel."""
        inlet_hl_max = (self.sed_tank_outlet_man_hl + self.sed_tank_diffuser_hl) * \
            (1 - self.SED_TANK_Q_RATIO ** 2)
        return inlet_hl_max

    @property
    def _inlet_w_pre_weir_plumbing_min(self):
        """Minimum width of the inlet channel (pre-weir) to fit pipes."""
        inlet_w_pre_weir_plumbing_min = pipe.fitting_od(self.sed_tank_inlet_man_nd) + \
            2 * self.fitting_s
        return inlet_w_pre_weir_plumbing_min

    @property
    def _inlet_w_pre_weir_hl_min(self):
        """Minimum width of the inlet channel (pre-weir) that doesn't exceed
        the permissible head loss.
        """
        inlet_w_pre_weir_hl_min = pc.horiz_chan_w(
            self.q,
            self.inlet_depth_max,
            self.inlet_hl_max,
            self.l,
            pc.viscosity_kinematic_water(self.temp),
            mat.CONCRETE_PIPE_ROUGH,
            False,
            0
        )
        return inlet_w_pre_weir_hl_min

    @property
    def inlet_w_pre_weir(self):
        """Width of the inlet channel (pre-weir)."""
        inlet_w_pre_weir = max(
            self._inlet_w_pre_weir_plumbing_min,
            self._inlet_w_pre_weir_hl_min)
        return inlet_w_pre_weir

    @property
    def _inlet_depth_plumbing_min(self):
        """Minimum depth of the inlet channel to fit pipes."""
        inlet_plumbing_depth_min = self.sed_tank_outlet_man_hl + self.sed_tank_diffuser_hl + \
            pipe.fitting_od(self.sed_tank_outlet_man_nd) + \
                self.fitting_s + self.outlet_weir_hl
        return inlet_plumbing_depth_min

    @property
    def _inlet_depth_hl_min(self):
        """Minimum depth of the inlet channel to stay within acceptable head
        loss.
        """
        inlet_chan_hl_depth = pc.horiz_chan_h(
            self.q,
            self.inlet_w_pre_weir,
            self.inlet_hl_max,
            self.l,
            pc.viscosity_kinematic_water(self.temp),
            mat.CONCRETE_PIPE_ROUGH,
            False)
        return inlet_chan_hl_depth

    @property
    def inlet_depth(self):
        """Depth of the inlet channel."""
        inlet_depth = max(self._inlet_depth_plumbing_min, self._inlet_depth_hl_min)
        return inlet_depth

    @property
    def inlet_weir_hl(self):
        """Head loss through the inlet channel weir."""
        inlet_weir_hl = pc.headloss_weir_rect(self.q, self.l)
        return inlet_weir_hl

    @property
    def inlet_h(self):
        """Height of the inlet channel."""
        inlet_h = self.inlet_depth + self.PLANT_FREEBOARD_H
        return inlet_h

    @property
    def inlet_weir_h(self):
        """Height of the inlet channel weir."""
        inlet_chan_h_weir = self.inlet_depth + self.WEIR_FREEBOARD_H
        return inlet_chan_h_weir

    @property
    def inlet_w_post_weir(self):
        """Width of the inlet channel (post-weir)"""
        inlet_w_post_weir = max(
            self.w_min,
            pc.horiz_chan_w(
                self.q,
                self.inlet_h,
                self.inlet_h,
                self.l,
                pc.viscosity_kinematic_water(self.temp),
                mat.CONCRETE_PIPE_ROUGH,
                1,
                0 ))
        return inlet_w_post_weir

    @property
    def inlet_w(self):
        """Width of the inlet channel"""
        inlet_w = self.inlet_w_pre_weir + self.weir_thickness + self.inlet_w_post_weir
        return inlet_w

    def _set_drain_pipe(self):
        drain_k_minor = hl.PIPE_ENTRANCE_K_MINOR + hl.PIPE_EXIT_K_MINOR + hl.EL90_K_MINOR
        drain_nd = pc.pipe_flow_nd(
            self.q,
            ut.get_sdr(self.drain_spec),
            self.SED_DEPTH_EST,
            self.SED_DEPTH_EST + self.inlet_w,
            pc.viscosity_kinematic_water(self.temp),
            mat.PVC_PIPE_ROUGH,
            drain_k_minor
        )

        self.drain_pipe = Pipe(
            size = drain_nd,
            spec = self.drain_spec,
            k_minor = drain_k_minor,
            )

    @property
    def inlet_drain_box_w(self):
        """Width of the inlet channel drain box"""
        inlet_drain_box_w = max( 2 * self.fitting_s + pipe.fitting_od(self.drain_pipe.size), self.inlet_w_post_weir)
        return inlet_drain_box_w

    @property
    def outlet_depth(self):
        """Depth of the outlet channel."""
        outlet_depth = self.inlet_depth - self.sed_tank_outlet_man_hl - \
           self.sed_tank_diffuser_hl
        return outlet_depth

    @property
    def outlet_weir_depth(self):
        """Depth of the outlet channel weir."""
        outlet_weir_depth = self.outlet_depth - self.weir_hl - self.WEIR_FREEBOARD_H
        return outlet_weir_depth

    @property
    def outlet_w_pre_weir(self):
        """Width of the outlet channel (pre-weir)."""
        return self.w_min

    @property
    def outlet_pipe_k_minor(self):
        outlet_pipe_k_minor = 2 * hl.EL90_K_MINOR + hl.PIPE_ENTRANCE_K_MINOR + \
                 hl.PIPE_EXIT_K_MINOR
        return outlet_pipe_k_minor

    @property
    def outlet_pipe_l(self):
        outlet_pipe_l = ha.DRAIN_CHAN_WALKWAY_W + self.inlet_w + 1.0 * u.m
        return outlet_pipe_l

    @property
    def outlet_pipe_q_max(self):
        """Maximum flow through the outlet pipe."""
        outlet_pipe_q_max = pc.flow_pipe(
            pipe.ID_SDR(self.outlet_pipe_nd_max, ut.get_sdr(self.outlet_pipe_spec)),
            self.outlet_pipe_hl_max,
            self.outlet_pipe_l,
            pc.viscosity_kinematic_water(self.temp),
            mat.PVC_PIPE_ROUGH,
            self.outlet_pipe_k_minor
        )
        return ut.round_step(
            outlet_pipe_q_max.to(u.L / u.s),
            step = 0.0001 * u.L / u.s
        )

    def _set_outlet_pipe(self):
        outlet_pipe_q = self.q / self.outlet_pipe_n

        # outlet_pipe_nd = pc.pipe_flow_nd(
        #     outlet_pipe_q,
        #     ut.get_sdr(self.outlet_pipe_spec),
        #     self.outlet_pipe_hl_max,
        #     self.outlet_pipe_l,
        #     pc.viscosity_kinematic_water(self.temp),
        #     mat.PVC_PIPE_ROUGH,
        #     self.outlet_pipe_k_minor
        # )

        outlet_pipe_nd = pc.pipe_flow_nd(
            outlet_pipe_q,
            ut.get_sdr(self.outlet_pipe_spec),
            self.outlet_pipe_hl_max,
            self.outlet_pipe_l,
            pc.viscosity_kinematic_water(self.temp),
            mat.PVC_PIPE_ROUGH,
            2 * hl.EL90_K_MINOR + hl.PIPE_ENTRANCE_K_MINOR + \
                 hl.PIPE_EXIT_K_MINOR
        )

        self.outlet_pipe = Pipe(
            l = self.outlet_pipe_l,
            q = outlet_pipe_q,
            size = outlet_pipe_nd,
            spec = self.outlet_pipe_spec,
            k_minor = self.outlet_pipe_k_minor
            )

    @property
    def outlet_pipe_n(self):
        """Number of outlet pipes."""
        outlet_pipe_n = math.ceil(self.q / self.outlet_pipe_q_max)
        return outlet_pipe_n

    @property
    def outlet_post_weir_w(self):
        """Width of the outlet channel (post-weir)."""
        outlet_post_weir_w = max(
            #need self.outlet_to_filter_nd
            self.fitting_s + pipe.fitting_od(self.outlet_pipe.size),
            self.fitting_s + pipe.fitting_od(self.drain_pipe.size),
            self.w_min,
            pc.horiz_chan_w(
                self.q,
                self.outlet_weir_depth - self.outlet_free_h, #what is outlet_free_h
                self.outlet_weir_depth,
                self.l,
                pc.viscosity_kinematic_water(self.temp),
                mat.PVC_PIPE_ROUGH,
                1,
                hl.PIPE_ENTRANCE_K_MINOR + hl.PIPE_EXIT_K_MINOR + hl.EL90_K_MINOR
            )
        )
        return outlet_post_weir_w

    @property
    def outlet_w(self):
        """Width of the outlet channel."""
        outlet_w = self.outlet_w_pre_weir + self.weir_thickness + \
            self.outlet_post_weir_w
        return outlet_w

    @property
    def outlet_drain_box_w(self):
        """Width of the outlet channel drain box."""
        outlet_drain_box_w = max(
            self.fitting_s + pipe.fitting_od(self.drain_pipe.size),
            self.outlet_post_weir_w
            )
        return outlet_drain_box_w

    @property
    def outlet_weir_h(self):
        """Height of the outlet channel weir."""
        outlet_weir_h = self.outlet_weir_depth + self.WEIR_FREEBOARD_H
        return outlet_weir_h

    @property
    def w_outer(self):
        """Outer width of the sedimentation channel."""
        w_outer = self.outlet_w + 2 * self.weir_thickness + self.inlet_w + self.sed_wall_thickness
        return w_outer

    @property
    def inlet_last_coupling_h(self):
        """Height of the last coupling in the inlet channel."""
        last_coupling_h = self.outlet_weir_depth - 2 * u.cm
        return last_coupling_h

    @property
    def inlet_step_h(self):
        """Height of the steps between each pipe in the inlet channel."""
        step_h = self.inlet_last_coupling_h / max(1, self.sed_tank_n - 1)
        return step_h

    @property
    def inlet_slope_l(self):
        """Length of the slopes between each pipe in the inlet channel."""
        inlet_slope_l = self.l + self.sed_wall_thickness - \
            pipe.fitting_od(self.sed_tank_inlet_man_nd) - self.fitting_s
        return inlet_slope_l



from aguaclara.design.component import Component

from aguaclara.design.ent_floc import EntTankFloc
from aguaclara.design.floc import Flocculator
from aguaclara.design.ent import EntranceTank
from aguaclara.design.lfom import LFOM

from aguaclara.design.sed import Sedimentor
from aguaclara.design.sed_tank import SedimentationTank
from aguaclara.design.sed_chan import SedimentationChannel

from aguaclara.core.units import u


class Plant(Component):
    """Functions for designing an AguaClara water treatment plant."""
    def __init__(self, **kwargs):
        self.etf = EntTankFloc()
        self.sed = Sedimentor()
        self.subcomponents = [self.etf, self.sed]

        super().__init__(**kwargs)
        super().set_subcomponents()

        self.design_floc()

    def design_floc(self):
        self.etf.floc.sed_chan_inlet_w_pre_weir = self.sed.chan.inlet_w_pre_weir


"""Includes all minor loss coefficient calculations for common complex
geometries, based off of the following resource:

https://neutrium.net/fluid_flow/pressure-loss-from-fittings-expansion-and-reduction-in-pipe-size/
"""

import aguaclara.core.physchem as pc
import aguaclara.core.constants as con
import aguaclara.core.materials as mats
import aguaclara.core.utility as ut
from aguaclara.core.units import u

import numpy as np


# TODO: Add units to docstrings. - Oliver Leung (oal22)


@ut.list_handler()
def k_value_expansion(ent_pipe_id, exit_pipe_id, q,
                      fitting_angle=180, rounded=False,
                      nu=con.WATER_NU, pipe_rough=mats.PVC_PIPE_ROUGH):
    """Calculates the minor loss coefficient (k-value) of a square,
    tapered, or rounded expansion in a pipe. Defaults to square.

    To use tapered, set angle to something that isn't 180.

    To use rounded, set rounded to True.

    Parameters:
        ent_pipe_id: Entrance pipe's inner diameter from which fluid flows.
        exit_pipe_id: Exit pipe's inner diameter to which fluid flows.
        q: Fluid's flow rate.

        fitting_angle: Fitting angle. Default: square (180 degrees).
        rounded: Rounded fitting. Default: square (False).

        nu: Fluid's dynamic viscosity of the fluid. Default: room
            temperature water (1 * 10**-6 * m**2/s)
        pipe_rough: Pipe roughness. Default: PVC pipe roughness

    Returns:
        k-value of expansion.
    """

    if ent_pipe_id > exit_pipe_id:
        print('k_value_expansion: Entrance pipe\'s inner diameter is larger '
              'than exit pipe\'s inner diameter, using reduction instead.')
        return k_value_reduction(ent_pipe_id, exit_pipe_id, q,
                                 fitting_angle, rounded,
                                 nu, pipe_rough)

    f = pc.fric_pipe(q, ent_pipe_id, nu, pipe_rough)  # Darcy friction factor.
    re = pc.re_pipe(q, ent_pipe_id, nu)          # Entrance pipe's Reynolds number.

    fitting_type = _get_fitting_type(fitting_angle, rounded)

    if fitting_type == 'square':
        result = _k_value_square_expansion(ent_pipe_id, exit_pipe_id, re, f)
    elif fitting_type == 'tapered':
        result = _k_value_tapered_expansion(ent_pipe_id, exit_pipe_id, re, f)
    elif fitting_type == 'rounded':
        result = _k_value_rounded_expansion(ent_pipe_id, exit_pipe_id, re)
    elif fitting_type == 'ambiguous':
        result = ValueError('The fitting is ambiguously both tapered and rounded. '
                         'Please set only either fitting_angle or rounded.')
    return result.to(u.dimensionless)


@ut.list_handler()
def k_value_reduction(ent_pipe_id, exit_pipe_id, q,
                      fitting_angle=180, rounded=False,
                      nu=con.WATER_NU, pipe_rough=mats.PVC_PIPE_ROUGH):
    """Calculates the minor loss coefficient (k-value) of a square,
    tapered, or rounded reduction in a pipe. Defaults to square.

    To use tapered, set angle to something that isn't 180.

    To use rounded, set rounded to True.

    Parameters:
        ent_pipe_id: Entrance pipe's inner diameter from which fluid flows.
        exit_pipe_id: Exit pipe's inner diameter to which fluid flows.
        q: Fluid's q rate.

        fitting_angle: Fitting angle. Default: square (180 degrees).
        rounded: Rounded fitting. Default: square (False).

        nu: Fluid's dynamic viscosity of the fluid. Default: room
            temperature water (1 * 10**-6 * m**2/s)
        pipe_rough: Pipe roughness. Default: PVC pipe roughness

    Returns:
        k-value of reduction.
    """

    if ent_pipe_id < exit_pipe_id:
        print('k_value_reduction: Entrance pipe\'s inner diameter is less than '
              'exit pipe\'s inner diameter, using expansion instead.')
        return k_value_expansion(ent_pipe_id, exit_pipe_id, q,
                                 fitting_angle, rounded,
                                 nu, pipe_rough)

    f = pc.fric_pipe(q, ent_pipe_id, nu, pipe_rough)     # Darcy friction factor.
    re = pc.re_pipe(q, ent_pipe_id, nu)             # Entrance pipe's Reynolds number.

    fitting_type = _get_fitting_type(fitting_angle, rounded)

    if fitting_type == 'square':
        result = _k_value_square_reduction(ent_pipe_id, exit_pipe_id, re, f)
    elif fitting_type == 'tapered':
        result = _k_value_tapered_reduction(ent_pipe_id, exit_pipe_id, fitting_angle, re, f)
    elif fitting_type == 'rounded':
        result = _k_value_rounded_reduction(ent_pipe_id, exit_pipe_id, re)
    elif fitting_type == 'ambiguous':
        raise ValueError('The fitting is ambiguously both tapered and rounded.'
                         'Please set only either fitting_angle or rounded.')
    return result.to(u.dimensionless)


@ut.list_handler()
def k_value_orifice(pipe_id, orifice_id, orifice_l, q,
                    nu=con.WATER_NU):
    """Calculates the minor loss coefficient of an orifice plate in a
    pipe.

    Parameters:
        pipe_id: Entrance pipe's inner diameter from which fluid flows.
        orifice_id: Orifice's inner diameter.
        orifice_l: Orifice's length from start to end.
        q: Fluid's q rate.

        nu: Fluid's dynamic viscosity of the fluid. Default: room
            temperature water (1 * 10**-6 * m**2/s)

    Returns:
        k-value at the orifice.
    """

    if orifice_id > pipe_id:
        raise ValueError('The orifice\'s inner diameter cannot be larger than'
                         'that of the entrance pipe.')

    re = pc.re_pipe(q, pipe_id, nu)  # Entrance pipe's Reynolds number.

    orifice_type = _get_orifice_type(orifice_l, orifice_id)

    if orifice_type == 'thin':
        result = _k_value_thin_sharp_orifice(pipe_id, orifice_id, re)
    elif orifice_type == 'thick':
        result = _k_value_thick_orifice(pipe_id, orifice_id, orifice_l, re)
    elif orifice_type == 'oversize':
        result = k_value_reduction(pipe_id, orifice_id, q) \
               + k_value_expansion(orifice_id, pipe_id, q)
    return result.to(u.dimensionless)


def _k_value_square_reduction(ent_pipe_id, exit_pipe_id, re, f):
    """Returns the minor loss coefficient for a square reducer.

    Parameters:
        ent_pipe_id: Entrance pipe's inner diameter.
        exit_pipe_id: Exit pipe's inner diameter.
        re: Reynold's number.
        f: Darcy friction factor.
    """

    if re < 2500:
        return (1.2 + (160 / re)) * ((ent_pipe_id / exit_pipe_id) ** 4)
    else:
        return (0.6 + 0.48 * f) * (ent_pipe_id / exit_pipe_id) ** 2\
            * ((ent_pipe_id / exit_pipe_id) ** 2 - 1)


def _k_value_tapered_reduction(ent_pipe_id, exit_pipe_id, fitting_angle, re, f):
    """Returns the minor loss coefficient for a tapered reducer.

    Parameters:
        ent_pipe_id: Entrance pipe's inner diameter.
        exit_pipe_id: Exit pipe's inner diameter.
        fitting_angle: Fitting angle between entrance and exit pipes.
        re: Reynold's number.
        f: Darcy friction factor.
    """

    k_value_square_reduction = _k_value_square_reduction(ent_pipe_id, exit_pipe_id,
                                                         re, f)

    if 45 < fitting_angle <= 180:
        return k_value_square_reduction * np.sqrt(np.sin(fitting_angle / 2))
    elif 0 < fitting_angle <= 45:
        return k_value_square_reduction * 1.6 * np.sin(fitting_angle / 2)
    else:
        raise ValueError('k_value_tapered_reduction: The reducer angle ('
                         + fitting_angle + ') cannot be outside of [0,180].')


def _k_value_rounded_reduction(id_entrance, id_exit, re):
    return (0.1 + (50 / re)) * ((id_entrance / id_exit) ** 4 - 1)


######### Expansions:


def _k_value_square_expansion(id_entrance: float, id_exit: float, re: float, f: float) -> float:
    # Calculate minor loss coefficient for square expansion
    if re < 4000:
        k = 2 * (1 - (id_entrance / id_exit) ** 4)
    else:
        k = (1 + 0.8 * f) * (1 - (id_entrance / id_exit) ** 2) ** 2
    return k


def _k_value_tapered_expansion(id_entrance: float, id_exit: float, re: float, f, theta: float) -> float:
    # Calculate minor loss coefficient for a tapered expansion
    k_square = _k_value_square_expansion(id_entrance, id_exit, re, f)
    if 45 < theta <= 180:
        k = k_square
    elif 0 < theta <= 45:
        k = k_square * 2.6 * np.sin(theta / 2)
    else:
        raise ValueError('The reducer angle cannot be outside the [0,180] range')
    return k


def _k_value_rounded_expansion(id_entrance: float, id_exit: float, re: float, f: float) -> float:
    return _k_value_square_expansion(id_entrance, id_exit, re, f)


######### Orifices:


def _k_value_thin_sharp_orifice(id_pipe: float, id_orifice: float, re: float) -> float:
    # Calculate minor loss coefficient for a thin, sharp orifice
    if re < 2500:
        k = ((2.72 + (id_orifice / id_pipe) ** 2) * ((120 / re) - 1)) * (1 - (id_orifice / id_pipe) ** 2) * \
            ((id_pipe / id_orifice) ** 4 - 1)
    else:
        k = ((2.72 + (id_orifice / id_pipe) ** 2) * (4000 / re)) * (1 - (id_orifice / id_pipe) ** 2) * \
            ((id_pipe / id_orifice) ** 4 - 1)
    return k


def _k_value_thick_orifice(id_pipe: float, id_orifice: float, length_orifice: float, re: float) -> float:
    # Calculate minor loss coefficient for a thick orifice
    # if L/id_orifice > 5, use the equation for a square reduction and expansion.
    k_thin = _k_value_thin_sharp_orifice(id_pipe, id_orifice, re)
    # thickness coefficient:
    c = 0.584 + (0.0936 / ((length_orifice / id_orifice) ** 1.5 + 0.225))
    return c * k_thin


def _get_fitting_type(fitting_angle, rounded):
    """Returns fitting type for expansions and reductions.

    Parameters:
        fitting_angle: Fitting angle. Usually is 180 for square fittings.
        rounded: Rounded fitting. Usually is False for square fittings.
    """

    if fitting_angle != 180 and rounded:
        return 'ambiguous'
    elif rounded:
        return 'rounded'
    elif fitting_angle != 180:
        return 'tapered'
    else:
        return 'square'


def _get_orifice_type(orifice_l, orifice_id):
    """Returns orifice type for orifice k-value caluclations.

    Parameters:
        orifice_l: Orifice's length.
        orifice_id: Orifice's inner diameter.
    """

    if orifice_l == 0:
        return 'thin'
    elif orifice_l / orifice_id < 5:
        return 'thick'
    else:
        return 'oversize'

#: 90 degree elbow
EL90_K_MINOR = 0.9

#:
EL45_K_MINOR = 0.45

#: The loss coefficient for the channel transition in a 90 degree turn
RIGHT_ANGLE_K_MINOR = 0.4

#:
ANGLE_VALVE_K_MINOR = 4.3

#:
GLOBE_VALVE_K_MINOR = 10

#:
GATE_VALVE_K_MINOR = 0.39

#:
CHECK_VALVE_CONV_K_MINOR = 4

#:
CHECK_VALVE_BALL_K_MINOR = 4.5

#: Headloss coefficient of a jet
EXP_K_MINOR = 1

#:
TEE_FLOW_RUN_K_MINOR = 0.6

#:
TEE_FLOW_BR_K_MINOR = 1.8

#:
PIPE_ENTRANCE_K_MINOR = 0.5

#:
PIPE_EXIT_K_MINOR = 1

#:
RM_GATE_VIN_K_MINOR = 25



"""The Linear Flow Orifice Meter (LFOM) of an AguaClara water treatment plant
imitates a sutro weir and produces a linear relation between flow rate and
water level within the entrance tank.

Example:
    >>> from aguaclara.core.units import u
    >>> from aguaclara.design.lfom import LFOM
    >>> lfom = LFOM(q=20 * u.L / u.s, hl=20 * u.cm)
    >>> lfom.row_n
    6
"""
import aguaclara.core.constants as con
import aguaclara.core.physchem as pc
import aguaclara.core.pipes as pipe
import aguaclara.core.utility as ut
import aguaclara.core.drills as drills
from aguaclara.core.units import u
from aguaclara.design.component import Component

import numpy as np
import math


class LFOM(Component):
    """Design and AguaClara plant's LFOM.

    Design Inputs:
        - ``q (float * u.L/u.s)``: Flow rate (recommended, defaults to 20L/s)
        - ``temp (float * u.degC)``: Water temperature (recommended, defaults to
          20°C)
        - ``hl (float * u.cm)``: Head loss (optional, defaults to 20cm)
        - ``safety_factor (float)``: Safety factor (optional, defaults to 1.5)
        - ``sdr (float)``: Standard dimension ratio (optional, defaults to 26)
        - ``drill_bits (float * u.inch array)``: List of drill bits
          (optional)
        - ``orifice_s (float * u.cm)``: The spacing between orifices (optional,
          defaults to 0.5cm)
        - ``min_row_n (int)``: Minimum number of rows of orifices (optional,
          defaults to 4)
        - ``max_row_n (int)``: Maximum number of rows of orifices (optional,
          defaults to 10)
    """
    def __init__(self, **kwargs):
        self.hl = 20.0 * u.cm
        self.safety_factor = 1.5
        self.sdr = 26.0
        self.drill_bits = drills.DRILL_BITS_D_IMPERIAL
        self.orifice_s = 0.5 * u.cm
        self.min_row_n = 4
        self.max_row_n = 10

        super().__init__(**kwargs)

    def stout_w_per_flow(self, h):
        """The width of a stout weir at a given elevation.

        Args:
            - ``h (float * u.m)``: Elevation height
        """
        w_per_flow = 2 / ((2 * u.gravity * h) ** (1 / 2) *
                          con.VC_ORIFICE_RATIO * np.pi * self.hl)
        return w_per_flow.to_base_units()

    @property
    def row_n(self):
        """ The number of rows."""
        N_estimated = (self.hl * np.pi / (2 * self.stout_w_per_flow(self.hl) * \
             self.q)).to(u.dimensionless)
        row_n = min(self.max_row_n,
                    max(self.min_row_n, math.trunc(N_estimated.magnitude)))
        return row_n

    @property
    def row_b(self):
        """The distance center to center between each row of orifices."""
        return self.hl / self.row_n

    @property
    def vel_critical(self):
        """The average vertical velocity of the water inside the LFOM pipe
        at the bottom of the orfices."""
        return (4 / (3 * math.pi) * (2 * u.gravity * self.hl) ** \
            (1 / 2)).to(u.m/u.s)

    @property
    def pipe_a_min(self):
        """The minimum cross-sectional area of the LFOM pipe assuring
        a safety factor."""
        return (self.safety_factor * self.q / self.vel_critical).to(u.cm**2)

    @property
    def pipe_nd(self):
        """The nominal diameter of the LFOM pipe"""
        ID = pc.diam_circle(self.pipe_a_min)
        return pipe.ND_SDR_available(ID, self.sdr)

    @property
    def top_row_orifice_a(self):
        """The orifice area corresponding to the top row of orifices."""
        z = self.hl - 0.5 * self.row_b
        return self.stout_w_per_flow(z) * self.q * self.row_b

    @property
    def orifice_d_max(self):
        """The maximum orifice diameter."""
        return pc.diam_circle(self.top_row_orifice_a)

    @property
    def orifice_d(self):
        """The actual orifice diameter."""
        maxdrill = min(self.row_b, self.orifice_d_max)
        return ut.floor_nearest(maxdrill, self.drill_bits)

    @property
    def drill_bit_a(self):
        """The area of the actual drill bit."""
        return pc.area_circle(self.orifice_d)

    @property
    def orifice_n_max_per_row(self):
        """The max number of orifices allowed in each row."""
        c = math.pi * pipe.ID_SDR(self.pipe_nd, self.sdr)
        b = self.orifice_d + self.orifice_s

        return math.floor(c/b)

    @property
    def q_per_row(self):
        """An array of flow at each row."""
        return np.linspace(1 / self.row_n, 1, self.row_n)*self.q

    @property
    def orifice_h_per_row(self):
        """The height of the center of each row of orifices."""
        height_orifices = (np.linspace(0, self.row_n - 1, self.row_n)) * \
            self.row_b + 0.5 * self.orifice_d
        return height_orifices

    def q_submerged(self, row_n, orifice_n_per_row):
        """The flow rate through some number of submerged rows.

        Args:
            - ``row_n``: Number of submerged rows
        """
        flow = 0 * u.L / u.s
        for i in range(row_n):
            flow = flow + (orifice_n_per_row[i] * (
               pc.flow_orifice_vert(self.orifice_d,
                                    self.row_b*(row_n + 1)
                                    - self.orifice_h_per_row[i],
                                    con.VC_ORIFICE_RATIO)))
        return flow.to(u.L / u.s)

    @property
    def orifice_n_per_row(self):
        """The number of orifices at each level."""
        h = self.row_b - 0.5*self.orifice_d
        flow_per_orifice = pc.flow_orifice_vert(self.orifice_d, h,
         con.VC_ORIFICE_RATIO)
        n = np.zeros(self.row_n)
        for i in range(self.row_n):
            flow_needed = self.q_per_row[i] - self.q_submerged(i, n)
            n_orifices_real = (flow_needed / \
                flow_per_orifice).to(u.dimensionless)
            n[i] = min((max(0, round(n_orifices_real))),
             self.orifice_n_max_per_row)
        return n

    @property
    def error_per_row(self):
        """The error of the design based off the predicted flow rate and
        the actual flow rate."""
        q_error = np.zeros(self.row_n)
        for i in range(self.row_n):
            actual_flow = self.q_submerged(i, self.orifice_n_per_row)
            q_error[i] = (((actual_flow - self.q_per_row[i]) / \
                 self.q_per_row[i]).to(u.dimensionless)).magnitude
        return q_error



"""
Tests for the research package's tube_sizing module.
"""

import unittest
from aguaclara.core.units import u
import aguaclara.research.peristaltic_pump as pp


class TestTubeSizing(unittest.TestCase):

    def assertAlmostEqualQuantity(self, first, second, places=7):
        self.assertAlmostEqual(first.magnitude, second.magnitude, places)
        self.assertAlmostEqual(first.units, second.units, places)

    def test_vol_per_rev_3_stop(self):
        self.assertAlmostEqualQuantity(0.0013286183895203283*u.mL/u.rev, pp.vol_per_rev_3_stop(color="orange-black"))
        self.assertAlmostEqualQuantity(0.14884596727278449*u.mL/u.rev, pp.vol_per_rev_3_stop(color="yellow-blue"))
        self.assertAlmostEqualQuantity(0.0031160704169596186*u.mL/u.rev, pp.vol_per_rev_3_stop(inner_diameter=.20*u.mm))
        self.assertAlmostEqualQuantity(0.4005495805189351*u.mL/u.rev, pp.vol_per_rev_3_stop(inner_diameter=2.79*u.mm))

    def test_ID_colored_tube(self):
        self.assertEqual(1.52*u.mm, pp.ID_colored_tube("yellow-blue"))
        self.assertEqual(0.51*u.mm, pp.ID_colored_tube("orange-yellow"))
        self.assertEqual(2.79*u.mm, pp.ID_colored_tube("purple-white"))

    def test_vol_per_rev_LS(self):
        self.assertEqual(0.06*u.mL/u.rev, pp.vol_per_rev_LS(13))
        self.assertEqual(1.6*u.mL/u.rev, pp.vol_per_rev_LS(15))
        self.assertEqual(3.8*u.mL/u.rev, pp.vol_per_rev_LS(18))
        self.assertEqual(4.8*u.mL/u.rev, pp.vol_per_rev_LS(36))

    def test_flow_rate(self):
        self.assertAlmostEqualQuantity(0.25*u.mL/u.s, pp.flow_rate(3*u.mL/u.rev, 5*u.rev/u.min))
        self.assertAlmostEqualQuantity(0.016666666666666666*u.mL/u.s, pp.flow_rate(.04*u.mL/u.rev, 25*u.rev/u.min))
        self.assertAlmostEqualQuantity(0.01*u.mL/u.s, pp.flow_rate(.001*u.mL/u.rev, 600*u.rev/u.min))



# -*- coding: utf-8 -*-
"""Lists of drill bit diameters."""
from aguaclara.core.units import u

import numpy as np


def get_drill_bits_d_imperial():
    """Return array of possible drill diameters in imperial."""
    step_32nd = np.arange(0.03125, 0.25, 0.03125)
    step_8th = np.arange(0.25, 1.0, 0.125)
    step_4th = np.arange(1.0, 2.0, 0.25)
    maximum = [2.0]

    return np.concatenate((step_32nd,
                           step_8th,
                           step_4th,
                           maximum)) * u.inch


def get_drill_bits_d_metric():
    """Return array of possible drill diameters in metric."""
    return np.concatenate((np.arange(1.0, 10.0, 0.1),
                           np.arange(10.0, 18.0, 0.5),
                           np.arange(18.0, 36.0, 1.0),
                           np.arange(40.0, 55.0, 5.0))) * u.mm


DRILL_BITS_D_IMPERIAL = get_drill_bits_d_imperial()
DRILL_BITS_D_METRIC = get_drill_bits_d_metric()



from setuptools import setup, find_packages

setup(
    name = 'aguaclara',
    version = '0.3.0',
    description = (
        'An open-source Python package for designing and performing research '
        'on AguaClara water treatment plants.'
    ),
    url = 'https://github.com/AguaClara/aguaclara',
    author = 'AguaClara Cornell',
    author_email = 'aguaclara@cornell.edu',
    license = 'MIT',
    packages = find_packages(),

    # Ensure that this matches Pipfile > [packages]
    install_requires = [
        'matplotlib',
        'urllib3',
        'pint',
        'pandas',
        'scipy',
        'onshape_client',
    ],

    python_requires='>=3.8',
    setup_requires=["pytest-runner"],
    tests_require=["pytest"],
    include_package_data=True,
    test_suite="tests",
    zip_safe=False
)
# test



"""The sedimentor of an AguaClara water treatment plant receives flocculated
water from the flocculator, then settles the flocs out via upflow and an array
of plate settlers.

Example:
    >>> from aguaclara.design.sed import *
    >>> sed = Sedimentor(q = 20 * u.L / u.s, temp = 20 * u.degC)
    >>> sed.tank_n
    4
"""
from aguaclara.design.sed_tank import SedimentationTank
from aguaclara.design.sed_chan import SedimentationChannel
from aguaclara.design.component import Component
from aguaclara.core.units import u
import aguaclara.core.constants as con

import numpy as np


class Sedimentor(Component):
    """Design an AguaClara sedimentor.

    The ``Sedimentor`` class designs the sedimentation tank and channel in
    tandem. For more information on those classes, see
    :class:`aguaclara.design.sed_tank.SedimentationTank` and
    :class:`aguaclara.design.sed_chan.SedimentationChannel`.

    Design inputs:
        - ``q (float * u.L / u.s)``: Flow rate (recommended, defaults to 20L/s)
        - ``temp (float * u.degC)``: Water temperature (recommended, defaults to
          20°C)
        - ``wall_thickness (float * u.cm)``: Wall thickness (optional, defaults
          to 15 * u.cm)
        - ``tank (SedimentationTank)``: Sedimentation Tank
          (optional, see :class:`aguaclara.design.sed_tank.SedimentationTank`
          for defaults)
        - ``chan (SedimentationChannel)``: Sedimentation Channel
          (optional, see
          :class:`aguaclara.design.sed_chan.SedimentationChannel` for defaults)
    """
    def __init__(self, **kwargs):
        self.wall_thickness = 15.0 * u.cm
        
        self.tank = SedimentationTank()
        self.chan = SedimentationChannel()
        self.subcomponents = [self.tank, self.chan]

        super().__init__(**kwargs)
        super().set_subcomponents()

        self._design_chan()
        self._design_tank()

    @property
    def tank_n(self):
        """The number of sedimentation tanks."""
        tank_n = np.ceil(self.q / self.tank.q_tank)
        return int(tank_n)
    
    def _design_chan(self):
        """Design the sedimentation channel based off of the tank."""
        self.chan.sed_tank_n = self.tank_n
        self.chan.sed_tank_diffuser_hl = self.tank.diffuser_hl
        self.chan.sed_tank_inlet_man_nd = self.tank.inlet_man_nd
        self.chan.sed_tank_outlet_man_nd = self.tank.outlet_man_nd
        self.chan.sed_tank_outlet_man_hl = self.tank.outlet_man_orifice_hl
        self.chan.sed_tank_diffuser_hl = self.tank.diffuser_hl
        self.chan.w_inner = self.tank.w_inner
        self.chan.sed_tank_wall_thickness = self.tank.WALL_THICKNESS
        self.chan.sed_wall_thickness = self.wall_thickness

    def _design_tank(self):
        """Design the sedimentation tank based off of the channel"""
        self.tank.sed_chan_w_outer = self.chan.w_outer
        self.tank.sed_chan_weir_thickness = self.chan.weir_thickness
        
MODULE_PLATES_N_MIN = 8

WATER_H_EST = 2 * u.m

GATE_VALVE_URL = "https://confluence.cornell.edu/download/attachments/173604905/Sed-Scaled-Gate-Valve-Threaded.dwg"

SUPPORT_BOLT_URL = "https://confluence.cornell.edu/download/attachments/173604905/PlateSettlerSupportBolt.dwg"

# ratio of the height to the width of the sedimentation tank inlet channel.
INLET_H_W_RATIO = 0.95

##############
# Exit launder
##############

# Center to center spacing of orifices in the launder
CENTER_LAUNDER_EST_DIST = 10 * u.cm

# The additional length needed in the launder cap pipe that is to be
# inserted into the launder coupling
LAUNDER_CAP_EXCESS_L = 3 * u.cm

# Space between the top of the plate settlers and the bottom of the
# launder pipe
LAMELLA_TO_LAUNDER_H = 5 * u.cm

# The additional length needed in the launder cap pipe that is to be
# inserted into the launder coupling

# Diameter of the pipe used to hold the plate settlers together
MOD_ND = 0.5 * u.inch

# Diameter of the pipe used to create spacers. The spacers slide over the
# 1/2" pipe and are between the plates
MOD_SPACER_ND = 0.75 * u.inch

MOD_SPACER_SDR = 17

# This is the vertical thickness of the lip where the lamella support sits. mrf222
LAMELLA_LEDGE_THICKNESS = 8 * u.cm

LAMELLA_PIPE_EDGE_S = 5 * u.cm

# Approximate x-dimension spacing between cross pipes in the plate settler
# support frame.
CENTER_PLATE_FRAME_CROSS_DIST_EST = 0.8 * u.m

# Estimated plate length used to get an initial estimate of sedimentation
# tank active length.
PLATE_L_EST = 60 * u.cm

# Pipe size of the support frame that holds up the plate settler modules
PLATE_FRAME_ND = 1.5 * u.inch

################
# Inlet manifold
################

# Max energy dissipation rate in the sed diffuser outletS
ENERGY_DIS_INT_MAX = 150 * u.mW/u.kg

# Ratio of min to max flow through the inlet manifold diffusers
MAN_ND_MAX = 8 * u.inch

# This is the minimum distance between the inlet manifold and the slope
# of the sed tank.
INLET_MAN_SLOPE_S = 10 * u.cm

# Length of exposed manifold stub coming out of the floc weir to which the
# free portion of the inlet manifold is attached with a flexible coupling.
MAN_CONNECTION_STUB_L = 4 * u.cm

# Space between the end of the manifold pipe and the edge of the first
# diffuser's hole, or the first manifold orifice.

MAN_FIRST_DIFFUSER_GAP_L = 3 * u.cm

#Vertical distance from the edge of the jet reverser half-pipe to the tip
# of the inlet manifold diffusers
JET_REVERSER_TO_DIFFUSERS_H = 3 * u.cm

#Gap between the end of the inlet manifold pipe and the end wall of the
# tank to be able to install the pipe
MAN_PIPE_FROM_TANK_END_L = 2  *u.cm

WALL_TO_DIFFUSER_GAP_L_MIN = 3 * u.cm

# Diameter of the holes drilled in the manifold so that the molded 1"
# diffuser pipes can fit tightly in place (normal OD of a 1" pipe is
# close to 1-5/16")
MAN_PORT_D = 1.25 * u.inch

# nominal diameter of pipe used for jet reverser in bottom of set tank

SDR_REVERSER = 26  # SDR of jet reverser pipe

DIFFUSER_PIPE_ND = 4 * u.cm  # nominal diameter of pipe used to make diffusers


# stretch factor applied to the diffuser PVC pipes as they are heated
# and molded
PVC_STRETCH_RATIO = 1.2

W_DIFFUSER_INNER = 0.3175 * u.cm  # opening width of diffusers

B_DIFFUSER = 5 * u.cm  # center to center spacing beteen diffusers

# Outlet to filter
# If the plant has two trains, the current design shows the exit channel
# continuing from one set of sed tanks into the filter inlet channel.
# The execution of this extended channel involves a few calculations.
FILTER_OUTLET_HL_MAX = 10 * u.cm

# Maximum length of sed plate sticking out past module pipes without any
# additional support. The goal is to prevent floppy modules that don't maintain
# constant distances between the plates

PLATE_CANTILEVERED_L = 20 * u.cm

HOPPER_DRAIN_ND = 1*u.inch

HOPPER_VIEWER_ND = 2*u.inch

HOPPER_SKIMMER_ND = 2*u.inch

# Diffusers/Jet Reverser

DIFFUSER_ND = 1*u.inch



"""The entrance tank of an AguaClara water treatment plant

#. removes large grit particles using plate settlers,
#. contains the :ref:`design-lfom`, which maintains a linear relation between flow and water level, and
#. introduces chemical dosing through the CDC <add link> using the water level set by the :ref:`design-lfom`.

Example:
    >>> from aguaclara.design.ent import *
    >>> ent_tank = EntranceTank(q = 20 * u.L / u.s, floc_chan_w = 42.0 * u.inch)
    >>> ent_tank.plate_n
    <Quantity(11.0, 'dimensionless')>
"""
import aguaclara.core.constants as con
import aguaclara.core.head_loss as hl
import aguaclara.core.materials as mat
import aguaclara.core.physchem as pc
import aguaclara.core.pipes as pipe
from aguaclara.core.units import u
import aguaclara.core.utility as ut

from aguaclara.design.component import Component
from aguaclara.design.pipeline import Pipe

import numpy as np


class EntranceTank(Component):
    """Design an AguaClara plant's entrance tank.

    An entrance tank's design relies on the LFOM's and flocculator's design in
    the same plant, but assumed/default values may be used to design an
    entrance tank by itself. To design these components in tandem, use
    :class:`aguaclara.design.ent_floc.EntTankFloc`.

    Design Inputs:
        - ``q (float * u.L / u.s)``: Flow rate (recommended, defaults to 20L/s)
        - ``temp (float * u.degC)``: Water temperature (recommended, defaults to
          20°C)
        - ``lfom_nd (float * u.inch)``: The LFOM's nominal diameter (recommended,
          defaults to 2")
        - ``floc_chan_w (float * u.inch)``: The flocculator's channel width
          (recommended, defaults to 42")
        - ``floc_chan_depth (float * u.m)``: The flocculator's channel depth
          (recommended, defaults to 2m)
        - ``plate_s (float * u.cm)``: The spacing between plates in a plate
          settler (optional, defaults to 2.5cm)
        - ``plate_thickness (float * u.deg)``: The thickness of a plate in a
          plate settler (optional, defaults to 2mm)
        - ``plate_angle (float * u.deg)``: The angle of the plate settler
          (optional, defaults to 60 degrees)
        - ``plate_capture_vel (float * u.mm / u.s)``: The capture velocity of the
          plate settler (optional, defaults to 8m/s)
        - ``fab_s(float * u.cm)``: The space needed for a person to remove
          the drain pipe (optional, defaults to 5cm)
        - ``sdr (float)``: Standard demension ratio (optional,
          defaults to 41)
    """
    def __init__(self, **kwargs):
        self.lfom_nd = 2.0 * u.inch # May be innacurate, check with Monroe -Oliver L., oal22, 4 Jun '19
        self.floc_chan_w = 42.0 * u.inch
        self.floc_end_depth = 2.0 * u.m
        self.plate_s = 2.5 * u.cm
        self.plate_thickness = 2.0 * u.mm
        self.plate_angle  =  50.0 * u.deg
        self.plate_capture_vel  =  8.0 * u.mm / u.s
        self.fab_s = 5.0 * u.cm
        self.spec = 'sdr41'

        self.drain_pipe = Pipe()
        self.subcomponents = [self.drain_pipe]

        super().__init__(**kwargs)
        self._set_drain_pipe()
        super().set_subcomponents()

    def _set_drain_pipe(self):
        """The inner diameter of the entrance tank drain pipe."""
        drain_pipe_k_minor = \
            hl.PIPE_ENTRANCE_K_MINOR + hl.PIPE_EXIT_K_MINOR + hl.EL90_K_MINOR

        nu = pc.viscosity_kinematic_water(self.temp)
        drain_id = pc.diam_pipe(self.q,
                                self.floc_end_depth,
                                self.floc_end_depth,
                                nu,
                                mat.PVC_PIPE_ROUGH,
                                drain_pipe_k_minor)

        self.drain_pipe = Pipe(
            id = drain_id,
            k_minor = drain_pipe_k_minor,
            spec = self.spec
        )

    @property
    def plate_n(self):
        """The number of plates in the plate settlers."""
        num_plates_as_float = \
            np.sqrt(
                (self.q / (
                    (self.plate_s + self.plate_thickness) * self.floc_chan_w *
                    self.plate_capture_vel *
                    np.sin(self.plate_angle.to(u.rad)).item()
                )).to(u.dimensionless)
            )
        num_plates_as_int = np.ceil(num_plates_as_float)
        return num_plates_as_int # This calculates to be too low. -Oliver

    @property
    def plate_l(self):
        """The length of the plates in the plate settlers."""
        plate_l = (
                self.q / (
                    self.plate_n * self.floc_chan_w * self.plate_capture_vel *
                    np.cos(self.plate_angle.to(u.rad))
                )
            ) - (self.plate_s * np.tan(self.plate_angle.to(u.rad)))
        plate_l_rounded = ut.ceil_step(plate_l.to(u.cm), 1.0 * u.cm)
        return plate_l_rounded

    @property
    def l(self):
        """The length of the entrance tank."""
        plate_array_thickness = \
            (self.plate_thickness * self.plate_n) + \
            (self.plate_s * (self.plate_n - 1))

        l = self.drain_pipe.od + (self.fab_s * 2) + \
            (
                plate_array_thickness * np.cos(((90 * u.deg) -
                self.plate_angle).to(u.rad))
            ) + \
            (self.plate_l * np.cos(self.plate_angle.to(u.rad))) + \
            (self.lfom_nd * 2)
        return l



"""This file contains functions which convert between the nominal, inner, and
outer diameters of pipes based on their standard dimension ratio (SDR).
"""
from aguaclara.core.units import u
import aguaclara.core.utility as ut
import numpy as np
import pandas as pd

import os.path
dir_path = os.path.dirname(__file__)
csv_path = os.path.join(dir_path, 'data/pipe_database.csv')
with open(csv_path) as pipedbfile:
    pipedb = pd.read_csv(pipedbfile)

# TODO: Add a deprecation warning for this once manifold design code has been
# implemented. The socket_depth and cap_thickness functions are used in
# a manifold calculation in sed_tank, and can also be transferred to pipeline
# design code once manifold design code has been implemented.

class Pipe:


    def __init__(self, nd,sdr):
        self.nd= nd
        self.sdr = sdr

    @property
    def od(self):
        index = (np.abs(np.array(pipedb['NDinch']) - self.nd.magnitude)).argmin()
        return pipedb.iloc[index, 1] * u.inch

    @property
    def id_sdr(self):
        return (self.od.magnitude * (self.sdr - 2) / self.sdr) * u.inch

    @property
    def id_sch40(self):
        myindex = (np.abs(np.array(pipedb['NDinch']) - self.nd.magnitude)).argmin()
        return (pipedb.iloc[myindex, 1] - 2 * (pipedb.iloc[myindex, 5])) * u.inch


@ut.list_handler()
def OD(ND):
    """Return a pipe's outer diameter according to its nominal diameter.

    :param ND: nominal diameter of pipe
    :type ND: u.inch

    :return: outer diameter of pipe, in inches
    :rtype: u.inch
    """
    # The pipe schedule is not required here because all of the pipes of a
    # given nominal diameter have the same outer diameter.
    #
    # Steps:
    # 1. Find the index of the closest nominal diameter.
    #    (Should this be changed to find the next largest ND?)
    # 2. Take the values of the array, subtract the ND, take the absolute
    #    value, find the index of the minimium value.
    ND = ND.to(u.inch).magnitude
    index = (np.abs(np.array(pipedb['NDinch']) - (ND))).argmin()
    return pipedb.iloc[index, 1] * u.inch


@ut.list_handler()
def fitting_od(pipe_nd, fitting_sdr=41):
    pipe_od = OD(pipe_nd)
    fitting_nd = ND_SDR_available(pipe_od, fitting_sdr)
    fitting_od = OD(fitting_nd)
    return fitting_od


@ut.list_handler()
def ID_SDR(ND, SDR):
    """Return the inner diameter of a pipe given its nominal diameter and SDR
    (standard diameter ratio).

    SDR is the outer diameter divided by the wall thickness.
    """
    return OD(ND) * (SDR-2) / SDR


@ut.list_handler()
def ID_sch40(ND):
    """Return the inner diameter for schedule 40 pipes.

    The wall thickness for these pipes is in the pipedb.

    Take the values of the array, subtract the ND, take the absolute
    value, find the index of the minimium value.
    """
    ND = ND.to(u.inch).magnitude
    myindex = (np.abs(np.array(pipedb['NDinch']) - (ND))).argmin()
    return (pipedb.iloc[myindex, 1] - 2*(pipedb.iloc[myindex, 5])) * u.inch


def ND_all_available():
    """Return an array of available nominal diameters.

    NDs available are those commonly used as based on the 'Used' column
    in the pipedb.
    """
    ND_all_available = []
    for i in range(len(pipedb['NDinch'])):
        if pipedb.iloc[i, 4] == 1:
            ND_all_available.append((pipedb['NDinch'][i]))
    return ND_all_available * u.inch


def od_all_available():
    """Return an array of available outer diameters.

    NDs available are those commonly used as based on the 'Used' column
    in the pipedb.
    """
    od_all_available = []
    for i in range(len(pipedb['ODinch'])):
        if pipedb.iloc[i, 4] == 1:
            od_all_available.append((pipedb['ODinch'][i]))
    return od_all_available * u.inch


@ut.list_handler()
def ID_SDR_all_available(SDR):
    """Return an array of inner diameters with a given SDR.

    IDs available are those commonly used based on the 'Used' column
    in the pipedb.
    """
    ID = []
    ND = ND_all_available()
    for i in range(len(ND)):
        ID.append(ID_SDR(ND[i], SDR).magnitude)
    return ID * u.inch


@ut.list_handler()
def ND_SDR_available(ID, SDR):
    """ Return an available ND given an ID and a schedule.

    Takes the values of the array, compares to the ID, and finds the index
    of the first value greater or equal.
    """
    for i in range(len(np.array(ID_SDR_all_available(SDR)))):
        if np.array(ID_SDR_all_available(SDR))[i] >= (ID.to(u.inch)).magnitude:
            return ND_all_available()[i]


@ut.list_handler()
def ND_available(NDguess):
    """Return the minimum ND that is available.

    1. Extract the magnitude in inches from the nominal diameter.
    2. Find the index of the closest nominal diameter.
    3. Take the values of the array, subtract the ND, take the
       absolute value, find the index of the minimium value.
    """
    myindex = (ND_all_available() >= NDguess)
    return min(ND_all_available()[myindex])


@ut.list_handler()
def od_available(od_guess):
    """Return the minimum OD that is available.

    1. Extract the magnitude in inches from the outer diameter.
    2. Find the index of the closest outer diameter.
    3. Take the values of the array, subtract the OD, take the
       absolute value, find the index of the minimium value.
    """
    myindex = (od_all_available() >= od_guess)
    return min(od_all_available()[myindex])


@ut.list_handler()
def socket_depth(nd):
    return nd / 2


@ut.list_handler()
def cap_thickness(nd):
    cap_thickness = (fitting_od(nd) - OD(ND_available(nd))) / 2
    return cap_thickness
